# Copyright (c) 2021 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_policy import policy as base_policy

from neutron import policy
from neutron.tests.unit.conf.policies import test_base as base


class FlavorAPITestCase(base.PolicyBaseTestCase):

    def setUp(self):
        super(FlavorAPITestCase, self).setUp()
        self.target = {'project_id': self.project_id}


class SystemAdminTests(FlavorAPITestCase):

    def setUp(self):
        super(SystemAdminTests, self).setUp()
        self.context = self.system_admin_ctx

    def test_create_flavor(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_flavor', self.target))

    def test_get_flavor(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_flavor', self.target))

    def test_update_flavor(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_flavor', self.target))

    def test_delete_flavor(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'delete_flavor', self.target))

    def test_create_service_profile(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_service_profile', self.target))

    def test_get_service_profile(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_service_profile', self.target))

    def test_update_service_profile(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_service_profile', self.target))

    def test_delete_service_profile(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'delete_service_profile', self.target))

    def test_create_flavor_service_profile(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_flavor_service_profile', self.target))

    def test_delete_flavor_service_profile(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'delete_flavor_service_profile', self.target))


class SystemMemberTests(FlavorAPITestCase):

    def setUp(self):
        super(SystemMemberTests, self).setUp()
        self.context = self.system_member_ctx

    def test_create_flavor(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_flavor', self.target)

    def test_get_flavor(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_flavor', self.target))

    def test_update_flavor(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_flavor', self.target)

    def test_delete_flavor(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_flavor', self.target)

    def test_create_service_profile(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_service_profile', self.target)

    def test_get_service_profile(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_service_profile', self.target))

    def test_update_service_profile(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_service_profile', self.target)

    def test_delete_service_profile(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_service_profile', self.target)

    def test_create_flavor_service_profile(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_flavor_service_profile',
            self.target)

    def test_delete_flavor_service_profile(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_flavor_service_profile',
            self.target)


class SystemReaderTests(SystemMemberTests):

    def setUp(self):
        super(SystemReaderTests, self).setUp()
        self.context = self.system_reader_ctx


class ProjectAdminTests(FlavorAPITestCase):

    def setUp(self):
        super(ProjectAdminTests, self).setUp()
        self.context = self.project_admin_ctx

    def test_create_flavor(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_flavor', self.target)

    def test_update_flavor(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_flavor', self.target)

    def test_delete_flavor(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'delete_flavor', self.target)

    def test_create_service_profile(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_service_profile', self.target)

    def test_update_service_profile(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_service_profile', self.target)

    def test_delete_service_profile(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'delete_service_profile', self.target)

    def test_create_flavor_service_profile(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_flavor_service_profile',
            self.target)

    def test_delete_flavor_service_profile(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'delete_flavor_service_profile',
            self.target)


class ProjectMemberTests(ProjectAdminTests):

    def setUp(self):
        super(ProjectMemberTests, self).setUp()
        self.context = self.project_member_ctx


class ProjectReaderTests(ProjectMemberTests):

    def setUp(self):
        super(ProjectReaderTests, self).setUp()
        self.context = self.project_reader_ctx



# Copyright (c) 2015 Taturiello Consulting, Meh.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import attributes
from neutron_lib.api import converters
from neutron_lib.db import constants as db_const
from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_utils import importutils
import pecan
from pecan import request

from neutron._i18n import _
from neutron.pecan_wsgi.controllers import utils
from neutron.quota import resource_registry

RESOURCE_NAME = "quota"
TENANT_ID_ATTR = {'tenant_id':
                  {'allow_post': False,
                   'allow_put': False,
                   'required_by_policy': True,
                   'validate': {'type:string': db_const.PROJECT_ID_FIELD_SIZE},
                   'is_visible': True}}


class QuotasController(utils.NeutronPecanController):

    def __init__(self):
        self._driver = importutils.import_class(
            cfg.CONF.QUOTAS.quota_driver
        )
        super(QuotasController, self).__init__(
            "%ss" % RESOURCE_NAME, RESOURCE_NAME)

    def _check_admin(self, context,
                     reason=_("Only admin can view or configure quota")):
        if not context.is_admin:
            raise n_exc.AdminRequired(reason=reason)

    @utils.expose()
    def _lookup(self, tenant_id, *remainder):
        return QuotaController(self._driver, tenant_id), remainder

    @utils.expose(generic=True)
    def index(self):
        neutron_context = request.context.get('neutron_context')
        # FIXME(salv-orlando): There shouldn't be any need to do this explicit
        # check. However some behaviours from the "old" extension have
        # been temporarily carried over here
        self._check_admin(neutron_context)
        # TODO(salv-orlando): proper plurals management
        return {self.collection:
                self._driver.get_all_quotas(
                    neutron_context,
                    resource_registry.get_all_resources())}

    @utils.when(index, method='POST')
    @utils.when(index, method='PUT')
    @utils.when(index, method='DELETE')
    def not_supported(self):
        pecan.abort(405)


class QuotaController(utils.NeutronPecanController):

    def __init__(self, _driver, tenant_id):
        self._driver = _driver
        self._tenant_id = tenant_id

        super(QuotaController, self).__init__(
            "%ss" % RESOURCE_NAME, RESOURCE_NAME)

        # Ensure limits for all registered resources are returned
        attr_dict = attributes.RESOURCES[self.collection]
        for quota_resource in resource_registry.get_all_resources().keys():
            attr_dict[quota_resource] = {
                'allow_post': False,
                'allow_put': True,
                'convert_to': converters.convert_to_int,
                'validate': {
                    'type:range': [-1, db_const.DB_INTEGER_MAX_VALUE]},
                'is_visible': True}
        # The quota resource must always declare a tenant_id attribute,
        # otherwise the attribute will be stripped off when generating the
        # response
        attr_dict.update(TENANT_ID_ATTR)

    @utils.expose(generic=True)
    def index(self):
        return get_tenant_quotas(self._tenant_id, self._driver)

    @utils.when(index, method='PUT')
    def put(self, *args, **kwargs):
        neutron_context = request.context.get('neutron_context')
        # For put requests there's always going to be a single element
        quota_data = request.context['resources'][0]
        for key, value in quota_data.items():
            self._driver.update_quota_limit(
                neutron_context, self._tenant_id, key, value)
        return get_tenant_quotas(self._tenant_id, self._driver)

    @utils.when_delete(index)
    def delete(self):
        neutron_context = request.context.get('neutron_context')
        self._driver.delete_tenant_quota(neutron_context,
                                         self._tenant_id)

    @utils.when(index, method='POST')
    def not_supported(self):
        pecan.abort(405)


def get_tenant_quotas(tenant_id, driver=None):
    if not driver:
        driver = importutils.import_class(cfg.CONF.QUOTAS.quota_driver)

    neutron_context = request.context.get('neutron_context')
    if tenant_id == 'tenant':
        # NOTE(salv-orlando): Read the following before the code in order
        # to avoid puking.
        # There is a weird undocumented behaviour of the Neutron quota API
        # as 'tenant' is used as an API action to return the identifier
        # of the tenant in the request context. This is used exclusively
        # for interaction with python-neutronclient and is a possibly
        # unnecessary 'whoami' API endpoint. Pending resolution of this
        # API issue, this controller will just treat the magic string
        # 'tenant' (and only that string) and return the response expected
        # by python-neutronclient
        return {'tenant': {'tenant_id': neutron_context.tenant_id}}
    tenant_quotas = driver.get_tenant_quotas(
        neutron_context,
        resource_registry.get_all_resources(),
        tenant_id)
    tenant_quotas['tenant_id'] = tenant_id
    return {RESOURCE_NAME: tenant_quotas}



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from neutron_lib.api.definitions import constants as api_const
from neutron_lib.callbacks import events
from neutron_lib import constants
from neutron_lib import context
from neutron_lib import exceptions as n_exc
from oslo_db import exception as db_exc
from oslo_utils import uuidutils
import testtools

from neutron.services.auto_allocate import db
from neutron.services.auto_allocate import exceptions
from neutron.tests.unit import testlib_api


DB_PLUGIN_KLASS = 'neutron.db.db_base_plugin_v2.NeutronDbPluginV2'


class AutoAllocateTestCase(testlib_api.SqlTestCase):

    def setUp(self):
        super(AutoAllocateTestCase, self).setUp()
        self.setup_coreplugin(core_plugin=DB_PLUGIN_KLASS)
        self.ctx = context.get_admin_context()
        self.mixin = db.AutoAllocatedTopologyMixin()
        self.mixin._l3_plugin = mock.Mock()
        self.mixin._core_plugin = mock.Mock()

    def test_ensure_external_network_default_value(self):
        network_id = uuidutils.generate_uuid()
        kwargs = {
            "context": self.ctx,
            "request": {
                "id": network_id,
                api_const.IS_DEFAULT: True
            },
            "network": {
                "id": network_id,
                api_const.IS_DEFAULT: False
            },
            "original_network": {
                "id": network_id,
                api_const.IS_DEFAULT: False
            }
        }
        network_mock = mock.MagicMock(network_id=network_id, is_default=False)
        with mock.patch(
            'neutron.objects.network.ExternalNetwork.get_objects',
            return_value=[network_mock]
        ) as get_external_nets, mock.patch(
            'neutron.objects.network.ExternalNetwork.get_object',
            return_value=network_mock
        ) as get_external_net:
            db._ensure_external_network_default_value_callback(
                "NETWORK", "precommit_update", "test_plugin",
                payload=events.DBEventPayload(
                    self.ctx, request_body=kwargs['request'],
                    states=(kwargs['original_network'], kwargs['network'])))
            get_external_nets.assert_called_once_with(
                self.ctx, _pager=mock.ANY, is_default=True)
            get_external_net.assert_called_once_with(
                self.ctx, network_id=network_id)
            network_mock.update.assert_called_once_with()

    def test_ensure_external_network_default_value_no_default(self):
        network_id = uuidutils.generate_uuid()
        kwargs = {
            "context": self.ctx,
            "request": {
                "id": network_id,
                api_const.IS_DEFAULT: None
            },
            "network": {
                "id": network_id,
                api_const.IS_DEFAULT: False
            },
        }
        network_mock = mock.MagicMock(network_id=network_id, is_default=False)
        with mock.patch(
            'neutron.objects.network.ExternalNetwork.get_objects',
            return_value=[network_mock]
        ) as get_external_nets, mock.patch(
            'neutron.objects.network.ExternalNetwork.get_object',
            return_value=network_mock
        ) as get_external_net:
            db._ensure_external_network_default_value_callback(
                "NETWORK", "precommit_update", "test_plugin",
                payload=events.DBEventPayload(
                    self.ctx, request_body=kwargs['request'],
                    states=(kwargs['network'],)))
            get_external_nets.assert_not_called()
            get_external_net.assert_not_called()
            network_mock.update.assert_not_called()

    def test_ensure_external_network_default_value_default_not_changed(self):
        network_id = uuidutils.generate_uuid()
        kwargs = {
            "context": self.ctx,
            "request": {
                "id": network_id,
                api_const.IS_DEFAULT: True
            },
            "network": {
                "id": network_id,
                api_const.IS_DEFAULT: True
            },
            "original_network": {
                "id": network_id,
                api_const.IS_DEFAULT: True
            }
        }
        network_mock = mock.MagicMock(network_id=network_id, is_default=True)
        with mock.patch(
            'neutron.objects.network.ExternalNetwork.get_objects',
            return_value=[network_mock]
        ) as get_external_nets, mock.patch(
            'neutron.objects.network.ExternalNetwork.get_object',
            return_value=network_mock
        ) as get_external_net:
            db._ensure_external_network_default_value_callback(
                "NETWORK", "precommit_update", "test_plugin",
                payload=events.DBEventPayload(
                    self.ctx, request_body=kwargs['request'],
                    states=(kwargs['original_network'], kwargs['network'])))
            get_external_nets.assert_called_once_with(
                self.ctx, _pager=mock.ANY, is_default=True)
            get_external_net.assert_not_called()
            network_mock.update.assert_not_called()

    def test_ensure_external_network_default_value_default_existed(self):
        network_id = uuidutils.generate_uuid()
        kwargs = {
            "context": self.ctx,
            "request": {
                "id": network_id,
                api_const.IS_DEFAULT: True
            },
            "network": {
                "id": network_id,
                api_const.IS_DEFAULT: False
            },
            "original_network": {
                "id": network_id,
                api_const.IS_DEFAULT: False
            }
        }
        network_mock = mock.MagicMock(network_id='fake_id', is_default=False)
        with mock.patch(
            'neutron.objects.network.ExternalNetwork.get_objects',
            return_value=[network_mock]
        ) as get_external_nets, mock.patch(
            'neutron.objects.network.ExternalNetwork.get_object',
            return_value=network_mock
        ) as get_external_net:
            self.assertRaises(exceptions.DefaultExternalNetworkExists,
                db._ensure_external_network_default_value_callback,
                "NETWORK", "precommit_update", "test_plugin",
                payload=events.DBEventPayload(
                    self.ctx, request_body=kwargs['request'],
                    states=(kwargs['original_network'], kwargs['network'])))
            get_external_nets.assert_called_once_with(
                self.ctx, _pager=mock.ANY, is_default=True)
            get_external_net.assert_not_called()
            network_mock.update.assert_not_called()

    def test__provision_external_connectivity_expected_cleanup(self):
        """Test that the right resources are cleaned up."""
        subnets = [
            {'id': 'subnet_foo_1', 'network_id': 'network_foo'},
            {'id': 'subnet_foo_2', 'network_id': 'network_foo'},
        ]
        with mock.patch.object(self.mixin, '_cleanup') as mock_cleanup:
            self.mixin.l3_plugin.create_router.return_value = (
                {'id': 'router_foo'})
            self.mixin.l3_plugin.add_router_interface.side_effect = (
                n_exc.BadRequest(resource='router', msg='doh!'))
            self.assertRaises(exceptions.AutoAllocationFailure,
                self.mixin._provision_external_connectivity,
                self.ctx, 'ext_net_foo', subnets, 'tenant_foo')
            # expect no subnets to be unplugged
            mock_cleanup.assert_called_once_with(
                self.ctx, network_id='network_foo',
                router_id='router_foo', subnets=[])

    def test__provision_external_connectivity_fail_expected_cleanup(self):
        """Test that the right resources are cleaned up."""
        subnets = [
            {'id': 'subnet_foo_1', 'network_id': 'network_foo'},
        ]
        with mock.patch.object(self.mixin, '_cleanup') as mock_cleanup:
            self.mixin.l3_plugin.create_router.side_effect = (
                n_exc.BadRequest(resource='router', msg='doh!'))
            self.assertRaises(exceptions.AutoAllocationFailure,
                self.mixin._provision_external_connectivity,
                self.ctx, 'ext_net_foo', subnets, 'tenant_foo')
            # expected router_id to be None
            mock_cleanup.assert_called_once_with(
                self.ctx, network_id='network_foo',
                router_id=None, subnets=[])

    def test_get_auto_allocated_topology_dry_run_happy_path_for_kevin(self):
        with mock.patch.object(self.mixin, '_check_requirements') as f:
            self.mixin.get_auto_allocated_topology(
                self.ctx, mock.ANY, fields=['dry-run'])
            self.assertEqual(1, f.call_count)

    def test_get_auto_allocated_topology_dry_run_bad_input(self):
        self.assertRaises(n_exc.BadRequest,
            self.mixin.get_auto_allocated_topology,
            self.ctx, mock.ANY, fields=['foo'])

    def test__provision_tenant_private_network_handles_subnet_errors(self):
        network_id = uuidutils.generate_uuid()
        self.mixin._core_plugin.create_network.return_value = (
            {'id': network_id})
        self.mixin._core_plugin.create_subnet.side_effect = (
            n_exc.SubnetAllocationError(reason='disaster'))
        with mock.patch.object(self.mixin, "_get_supported_subnetpools") as f,\
                mock.patch.object(self.mixin, "_cleanup") as g:
            f.return_value = (
                [{'ip_version': constants.IP_VERSION_4,
                  'id': uuidutils.generate_uuid()}])
            self.assertRaises(exceptions.AutoAllocationFailure,
                              self.mixin._provision_tenant_private_network,
                              self.ctx, 'foo_tenant')
            g.assert_called_once_with(self.ctx, network_id)

    def _test__build_topology(self, method, provisioning_exception):
        with mock.patch.object(self.mixin,
                               method,
                               side_effect=provisioning_exception), \
                mock.patch.object(self.mixin, '_cleanup') as f:
            self.assertRaises(provisioning_exception.error,
                              self.mixin._build_topology,
                              self.ctx, mock.ANY, 'foo_net')
            f.assert_called_once_with(
                self.ctx,
                network_id=provisioning_exception.network_id,
                router_id=provisioning_exception.router_id,
                subnets=provisioning_exception.subnets
            )

    def test__build_topology_provisioning_error_no_toplogy(self):
        provisioning_exception = exceptions.UnknownProvisioningError(
            db_exc.DBError)
        self._test__build_topology(
            '_provision_tenant_private_network',
            provisioning_exception)

    def test__build_topology_provisioning_error_network_only(self):
        provisioning_exception = exceptions.UnknownProvisioningError(
            Exception, network_id='foo')
        self._test__build_topology(
            '_provision_tenant_private_network',
            provisioning_exception)

    def test__build_topology_error_only_network_again(self):
        provisioning_exception = exceptions.UnknownProvisioningError(
            AttributeError, network_id='foo')
        with mock.patch.object(self.mixin,
                               '_provision_tenant_private_network') as f:
            f.return_value = [{'network_id': 'foo'}]
            self._test__build_topology(
                '_provision_external_connectivity',
                provisioning_exception)

    def test__build_topology_error_network_with_router(self):
        provisioning_exception = exceptions.UnknownProvisioningError(
            KeyError, network_id='foo_n', router_id='foo_r')
        with mock.patch.object(self.mixin,
                               '_provision_tenant_private_network') as f:
            f.return_value = [{'network_id': 'foo_n'}]
            self._test__build_topology(
                '_provision_external_connectivity',
                provisioning_exception)

    def test__build_topology_error_network_with_router_and_interfaces(self):
        provisioning_exception = exceptions.UnknownProvisioningError(
            db_exc.DBConnectionError,
            network_id='foo_n', router_id='foo_r', subnets=[{'id': 'foo_s'}])
        with mock.patch.object(self.mixin,
                               '_provision_tenant_private_network') as f,\
                mock.patch.object(self.mixin,
                                  '_provision_external_connectivity') as g:
            f.return_value = [{'network_id': 'foo_n'}]
            g.return_value = {'id': 'foo_r'}
            self._test__build_topology(
                '_save',
                provisioning_exception)

    def test__save_with_provisioning_error(self):
        self.mixin._core_plugin.update_network.side_effect = Exception
        with testtools.ExpectedException(
                exceptions.UnknownProvisioningError) as e:
            self.mixin._save(self.ctx, 'foo_t', 'foo_n', 'foo_r',
                             [{'id': 'foo_s'}])
            self.assertEqual('foo_n', e.network_id)
            self.assertEqual('foo_r', e.router_id)
            self.assertEqual([{'id': 'foo_s'}], e.subnets)

    def test__provision_external_connectivity_with_provisioning_error(self):
        self.mixin._l3_plugin.create_router.side_effect = Exception
        with testtools.ExpectedException(
                exceptions.UnknownProvisioningError) as e:
            self.mixin._provision_external_connectivity(
                self.ctx, 'foo_default',
                [{'id': 'foo_s', 'network_id': 'foo_n'}],
                'foo_tenant')
            self.assertEqual('foo_n', e.network_id)
            self.assertIsNone(e.router_id)
            self.assertIsNone(e.subnets)

    def test__provision_tenant_private_network_with_provisioning_error(self):
        self.mixin._core_plugin.create_network.side_effect = Exception
        with testtools.ExpectedException(
                exceptions.UnknownProvisioningError) as e:
            self.mixin._provision_tenant_private_network(
                self.ctx, 'foo_tenant')
            self.assertIsNone(e.network_id)

    def test__check_requirements_fail_on_missing_ext_net(self):
        self.assertRaises(exceptions.AutoAllocationFailure,
            self.mixin._check_requirements, self.ctx, 'foo_tenant')

    def test__check_requirements_fail_on_missing_pools(self):
        with mock.patch.object(
            self.mixin, '_get_default_external_network'),\
            mock.patch.object(
                self.mixin, '_get_supported_subnetpools') as g:
            g.side_effect = n_exc.NotFound()
            self.assertRaises(exceptions.AutoAllocationFailure,
                self.mixin._check_requirements, self.ctx, 'foo_tenant')

    def test__check_requirements_happy_path_for_kevin(self):
        with mock.patch.object(
            self.mixin, '_get_default_external_network'),\
            mock.patch.object(
                self.mixin, '_get_supported_subnetpools'):
            result = self.mixin._check_requirements(self.ctx, 'foo_tenant')
            expected = {'id': 'dry-run=pass', 'tenant_id': 'foo_tenant'}
            self.assertEqual(expected, result)

    def test__cleanup_handles_failures(self):
        notfound = n_exc.NotFound
        self.mixin._l3_plugin.remove_router_interface.side_effect = (
            notfound)
        self.mixin._l3_plugin.delete_router.side_effect = (
            notfound)
        self.mixin._core_plugin.delete_network.side_effect = (
            notfound)
        self.mixin._cleanup(self.ctx, network_id=44, router_id=45,
                            subnets=[{'id': 46}])



# Copyright (c) 2015 Mellanox Technologies, Ltd
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.agent import l3_extension
from neutron_lib import exceptions
from oslo_log import log

from neutron._i18n import _
from neutron.agent import agent_extensions_manager as agent_ext_manager
from neutron.conf.agent import agent_extensions_manager as agent_ext_mgr_config

LOG = log.getLogger(__name__)


L3_AGENT_EXT_MANAGER_NAMESPACE = 'neutron.agent.l3.extensions'


# TODO(ralonsoh): rehome to neutron_lib.
class L3ExtensionException(exceptions.NeutronException):
    message = _('The following L3 agent extensions do not inherit from '
                '``neutron_lib.agent.l3_extension.L3AgentExtension``: '
                '%(extensions)s.')


def register_opts(conf):
    agent_ext_mgr_config.register_agent_ext_manager_opts(conf)


class L3AgentExtensionsManager(agent_ext_manager.AgentExtensionsManager):
    """Manage l3 agent extensions."""

    def __init__(self, conf):
        super(L3AgentExtensionsManager,
              self).__init__(conf, L3_AGENT_EXT_MANAGER_NAMESPACE)
        extensions = []
        for extension in self:
            if not isinstance(extension.obj, (l3_extension.L3AgentExtension,)):
                extensions.append(extension.attr)
        if extensions:
            raise L3ExtensionException(extensions=extensions)

    def add_router(self, context, data):
        """Notify all agent extensions to add router."""
        for extension in self:
            extension.obj.add_router(context, data)
        LOG.debug("L3 agent extension(s) finished router %s "
                  "add action.", data['id'])

    def update_router(self, context, data):
        """Notify all agent extensions to update router."""
        for extension in self:
            extension.obj.update_router(context, data)
        LOG.debug("L3 agent extension(s) finished router %s "
                  "update action.", data['id'])

    def delete_router(self, context, data):
        """Notify all agent extensions to delete router."""
        for extension in self:
            extension.obj.delete_router(context, data)
        LOG.debug("L3 agent extension(s) finished router %s "
                  "delete action.", data['id'])

    def ha_state_change(self, context, data):
        """Notify all agent extensions for HA router state change."""
        for extension in self:
            extension.obj.ha_state_change(context, data)
        LOG.debug('L3 agent extension(s) finished HA state change to "%s" '
                  'for router %s', data['state'], data['router_id'])



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from oslo_log import versionutils
from oslo_policy import policy

from neutron.conf.policies import base


COLLECTION_PATH = '/floatingips'
RESOURCE_PATH = '/floatingips/{id}'

DEPRECATION_REASON = (
    "The Floating IP API now supports system scope and default roles.")

rules = [
    policy.DocumentedRuleDefault(
        name='create_floatingip',
        check_str=base.PROJECT_MEMBER,
        description='Create a floating IP',
        operations=[
            {
                'method': 'POST',
                'path': COLLECTION_PATH,
            },
        ],
        scope_types=['project'],
        deprecated_rule=policy.DeprecatedRule(
            name='create_floatingip',
            check_str=base.RULE_ANY,
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_floatingip:floating_ip_address',
        check_str=base.PROJECT_ADMIN,
        description='Create a floating IP with a specific IP address',
        operations=[
            {
                'method': 'POST',
                'path': COLLECTION_PATH,
            },
        ],
        scope_types=['project'],
        deprecated_rule=policy.DeprecatedRule(
            name='create_floatingip:floating_ip_address',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_floatingip',
        check_str=base.PROJECT_READER,
        description='Get a floating IP',
        operations=[
            {
                'method': 'GET',
                'path': COLLECTION_PATH,
            },
            {
                'method': 'GET',
                'path': RESOURCE_PATH,
            },
        ],
        scope_types=['project'],
        deprecated_rule=policy.DeprecatedRule(
            name='get_floatingip',
            check_str=base.RULE_ADMIN_OR_OWNER,
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_floatingip',
        check_str=base.PROJECT_MEMBER,
        description='Update a floating IP',
        operations=[
            {
                'method': 'PUT',
                'path': RESOURCE_PATH,
            },
        ],
        scope_types=['project'],
        deprecated_rule=policy.DeprecatedRule(
            name='update_floatingip',
            check_str=base.RULE_ADMIN_OR_OWNER,
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_floatingip',
        check_str=base.PROJECT_MEMBER,
        description='Delete a floating IP',
        operations=[
            {
                'method': 'DELETE',
                'path': RESOURCE_PATH,
            },
        ],
        scope_types=['project'],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_floatingip',
            check_str=base.RULE_ADMIN_OR_OWNER,
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
]


def list_rules():
    return rules



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron_lib import constants
from oslo_config import cfg

from neutron.services.trunk.drivers.linuxbridge import driver
from neutron.tests import base


class LinuxBridgeDriverTestCase(base.BaseTestCase):

    def test_driver_is_loaded(self):
        inst = driver.LinuxBridgeDriver.create()
        cfg.CONF.set_override('mechanism_drivers',
                              ['a', 'b', 'linuxbridge'], group='ml2')
        self.assertTrue(inst.is_loaded)
        cfg.CONF.set_override('mechanism_drivers',
                              ['a', 'b'], group='ml2')
        self.assertFalse(inst.is_loaded)
        cfg.CONF.set_override('core_plugin', 'my_foo_plugin')
        self.assertFalse(inst.is_loaded)

    def test_driver_properties(self):
        inst = driver.LinuxBridgeDriver.create()
        self.assertEqual(driver.NAME, inst.name)
        self.assertEqual(driver.SUPPORTED_INTERFACES, inst.interfaces)
        self.assertEqual(driver.SUPPORTED_SEGMENTATION_TYPES,
                         inst.segmentation_types)
        self.assertEqual(constants.AGENT_TYPE_LINUXBRIDGE, inst.agent_type)
        self.assertTrue(inst.can_trunk_bound_port)



# Copyright 2018 OpenStack Foundation
# Copyright 2017 Letv Cloud Computing
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

from neutron_lib.api.definitions import l3 as l3_apidef
from neutron_lib.api.definitions import qos_gateway_ip
from neutron_lib.api import extensions
from neutron_lib.db import api as db_api
from neutron_lib.db import resource_extend
from neutron_lib.services.qos import constants as qos_consts
from oslo_log import log as logging

from neutron.db import l3_db
from neutron.db import l3_gwmode_db
from neutron.objects.qos import policy as policy_object

LOG = logging.getLogger(__name__)


@resource_extend.has_resource_extenders
class L3_gw_ip_qos_dbonly_mixin(l3_gwmode_db.L3_NAT_dbonly_mixin):
    """Mixin class to add router gateway IP's QoS extra attributes."""

    _gw_ip_qos = None

    @staticmethod
    @resource_extend.extends([l3_apidef.ROUTERS])
    def _extend_router_dict_gw_qos(router_res, router_db):
        if router_db.gw_port_id and router_db.get('qos_policy_binding'):
            policy_id = router_db.qos_policy_binding.policy_id
            router_res[l3_apidef.EXTERNAL_GW_INFO].update(
                {qos_consts.QOS_POLICY_ID: policy_id})

    @property
    def _is_gw_ip_qos_supported(self):
        if self._gw_ip_qos is None:
            # Check L3 service plugin
            self._gw_ip_qos = extensions.is_extension_supported(
                self, qos_gateway_ip.ALIAS)
        return self._gw_ip_qos

    def _create_gw_ip_qos_db(self, context, router_id, policy_id):
        policy = policy_object.QosPolicy.get_policy_obj(context, policy_id)
        policy.attach_router(router_id)

    def _delete_gw_ip_qos_db(self, context, router_id, policy_id):
        policy = policy_object.QosPolicy.get_policy_obj(context, policy_id)
        policy.detach_router(router_id, if_exists=True)

    def _update_router_gw_info(self, context, router_id, info,
                               request_body, router=None):
        # Calls superclass, pass router db object for avoiding re-loading
        router = super(L3_gw_ip_qos_dbonly_mixin,
                       self)._update_router_gw_info(
            context, router_id, info, request_body, router)

        if not self._is_gw_ip_qos_supported:
            return router

        with db_api.CONTEXT_WRITER.using(context):
            if not router.gw_port:
                info = {qos_consts.QOS_POLICY_ID: None}
            self._update_router_gw_qos_policy(context, router_id, info, router)

        return router

    def _get_router_gateway_policy_binding(self, context, router_id):
        router = self._get_router(context, router_id)
        return router.qos_policy_binding

    def _update_router_gw_qos_policy(self, context, router_id, info, router):
        if not info or qos_consts.QOS_POLICY_ID not in info:
            # An explicit 'None' for `qos_policy_id` indicates to clear
            # the router gateway IP policy. So if info does not have
            # the key `qos_polcy_id`, we can not decide what behavior
            # to be done, then directly return here.
            return

        new_qos_policy_id = info[qos_consts.QOS_POLICY_ID]
        if router.qos_policy_binding:
            old_qos_policy_id = router.qos_policy_binding.policy_id

            if old_qos_policy_id == new_qos_policy_id:
                return
            if old_qos_policy_id:
                self._delete_gw_ip_qos_db(context,
                                          router_id,
                                          old_qos_policy_id)

        if new_qos_policy_id:
            self._create_gw_ip_qos_db(
                context, router_id, new_qos_policy_id)

    def _build_routers_list(self, context, routers, gw_ports):
        routers = super(L3_gw_ip_qos_dbonly_mixin,
                        self)._build_routers_list(
                            context, routers, gw_ports)

        if not self._is_gw_ip_qos_supported:
            return routers

        for rtr in routers:
            gw_port_id = rtr['gw_port_id']
            # Collect gw ports only if available
            if gw_port_id and gw_ports.get(gw_port_id):
                rtr['gw_port'] = gw_ports[gw_port_id]
                router_gateway_policy_binding = (
                    self._get_router_gateway_policy_binding(
                        context, rtr['id']))
                qos_policy_id = None
                if router_gateway_policy_binding:
                    qos_policy_id = router_gateway_policy_binding.policy_id
                rtr['gw_port'][qos_consts.QOS_POLICY_ID] = qos_policy_id
        return routers


class L3_gw_ip_qos_db_mixin(L3_gw_ip_qos_dbonly_mixin,
                            l3_db.L3_NAT_db_mixin):
    pass



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.db import model_base
from oslo_utils import timeutils
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite


class OVNRevisionNumbers(model_base.BASEV2):
    __tablename__ = 'ovn_revision_numbers'

    standard_attr_id = sa.Column(
        sa.BigInteger().with_variant(sa.Integer(), 'sqlite'),
        sa.ForeignKey('standardattributes.id', ondelete='SET NULL'),
        nullable=True)
    resource_uuid = sa.Column(sa.String(36), nullable=False, index=True)
    resource_type = sa.Column(sa.String(36), nullable=False, index=True)
    revision_number = sa.Column(
        sa.BigInteger().with_variant(sa.Integer(), 'sqlite'),
        server_default='0', default=0, nullable=False)
    created_at = sa.Column(
        sa.DateTime().with_variant(
            sqlite.DATETIME(truncate_microseconds=True), 'sqlite'),
        default=sa.func.now(), nullable=False)
    updated_at = sa.Column(sa.TIMESTAMP, default=sa.func.now(),
                           onupdate=sa.func.now(), nullable=True)

    __table_args__ = (
        sa.PrimaryKeyConstraint(
            resource_uuid, resource_type,
            name='ovn_revision_numbers0resource_uuid0resource_type'),
        model_base.BASEV2.__table_args__
    )


class OVNHashRing(model_base.BASEV2):
    __tablename__ = 'ovn_hash_ring'

    node_uuid = sa.Column(sa.String(36), nullable=False, index=True)
    group_name = sa.Column(sa.String(256), nullable=False, index=True)
    hostname = sa.Column(sa.String(256), nullable=False)
    created_at = sa.Column(sa.DateTime(), default=timeutils.utcnow,
                           nullable=False)
    updated_at = sa.Column(sa.DateTime(), default=timeutils.utcnow,
                           nullable=False)
    __table_args__ = (
        sa.PrimaryKeyConstraint(
            node_uuid, group_name,
            name='ovn_hash_ring0node_uuid0group_name'),
        model_base.BASEV2.__table_args__
    )



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging as sys_logging

from oslo_reports import guru_meditation_report as gmr

from neutron import version

# During the call to gmr.TextGuruMeditation.setup_autorun(), Guru Meditation
# Report tries to start logging. Set a handler here to accommodate this.
logger = sys_logging.getLogger(None)
if not logger.handlers:
    logger.addHandler(sys_logging.StreamHandler())

_version_string = version.version_info.release_string()
gmr.TextGuruMeditation.setup_autorun(version=_version_string)



# Copyright (c) 2015 Thales Services SAS
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
import sqlalchemy as sa

"""subnetpool hash

Revision ID: 26c371498592
Revises: 45f955889773
Create Date: 2015-06-02 21:18:19.942076

"""

# revision identifiers, used by Alembic.
revision = '26c371498592'
down_revision = '45f955889773'


def upgrade():
    op.add_column(
        'subnetpools',
        sa.Column('hash', sa.String(36), nullable=False, server_default=''))



# Copyright 2015 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import configparser
import os.path

import fixtures

from neutron.tests import base


class ConfigDict(base.AttributeDict):
    def update(self, other):
        self.convert_to_attr_dict(other)
        super(ConfigDict, self).update(other)

    def convert_to_attr_dict(self, other):
        """Convert nested dicts to AttributeDict.

        :param other: dictionary to be directly modified.
        """
        for key, value in other.items():
            if isinstance(value, dict):
                if not isinstance(value, base.AttributeDict):
                    other[key] = base.AttributeDict(value)
                self.convert_to_attr_dict(value)


class ConfigFileFixture(fixtures.Fixture):
    """A fixture that knows how to translate configurations to files.

    :param base_filename: the filename to use on disk.
    :param config: a ConfigDict instance.
    :param temp_dir: an existing temporary directory to use for storage.
    """

    def __init__(self, base_filename, config, temp_dir):
        super(ConfigFileFixture, self).__init__()
        self.base_filename = base_filename
        self.config = config
        self.temp_dir = temp_dir

    def _setUp(self):
        self.write_config_to_configfile()

    def write_config_to_configfile(self):
        config_parser = self.dict_to_config_parser(self.config)
        # Need to randomly generate a unique folder to put the file in
        self.filename = os.path.join(self.temp_dir, self.base_filename)
        with open(self.filename, 'w') as f:
            config_parser.write(f)
            f.flush()

    def dict_to_config_parser(self, config_dict):
        config_parser = configparser.ConfigParser()
        for section, section_dict in config_dict.items():
            if section != 'DEFAULT':
                config_parser.add_section(section)
            for option, value in section_dict.items():
                try:
                    config_parser.set(section, option, value)
                except TypeError as te:
                    raise TypeError(
                        "%(msg)s: section %(section)s, option %(option)s, "
                        "value: %(value)s" % {
                            'msg': te.args[0],
                            'section': section,
                            'option': option,
                            'value': value,
                        })
        return config_parser



# Copyright (c) 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
from datetime import datetime
import itertools

import netaddr
from neutron_lib.agent import topics
from neutron_lib.api.definitions import portbindings_extended as pb_ext
from neutron_lib.callbacks import events as callback_events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources as callback_resources
from neutron_lib import constants
from neutron_lib.plugins import utils
from neutron_lib import rpc as lib_rpc
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging
from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from neutron.agent import resource_cache
from neutron.api.rpc.callbacks import resources
from neutron import objects

LOG = logging.getLogger(__name__)
BINDING_DEACTIVATE = 'binding_deactivate'
DeviceInfo = collections.namedtuple('DeviceInfo', 'mac pci_slot')


def create_consumers(endpoints, prefix, topic_details, start_listening=True):
    """Create agent RPC consumers.

    :param endpoints: The list of endpoints to process the incoming messages.
    :param prefix: Common prefix for the plugin/agent message queues.
    :param topic_details: A list of topics. Each topic has a name, an
                          operation, and an optional host param keying the
                          subscription to topic.host for plugin calls.
    :param start_listening: if True, it starts the processing loop

    :returns: A common Connection.
    """

    connection = lib_rpc.Connection()
    for details in topic_details:
        topic, operation, node_name = itertools.islice(
            itertools.chain(details, [None]), 3)

        topic_name = topics.get_topic_name(prefix, topic, operation)
        connection.create_consumer(topic_name, endpoints, fanout=True)
        if node_name:
            node_topic_name = '%s.%s' % (topic_name, node_name)
            connection.create_consumer(node_topic_name,
                                       endpoints,
                                       fanout=False)
    if start_listening:
        connection.consume_in_threads()
    return connection


class PluginReportStateAPI(object):
    """RPC client used to report state back to plugin.

    This class implements the client side of an rpc interface.  The server side
    can be found in neutron.db.agents_db.AgentExtRpcCallback.  For more
    information on changing rpc interfaces, see
    doc/source/contributor/internals/rpc_api.rst.
    """
    def __init__(self, topic):
        target = oslo_messaging.Target(topic=topic, version='1.2',
                                       namespace=constants.RPC_NAMESPACE_STATE)
        self.client = lib_rpc.get_client(target)
        self.timeout = cfg.CONF.AGENT.report_interval

    def has_alive_neutron_server(self, context, **kwargs):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'has_alive_neutron_server', **kwargs)

    def report_state(self, context, agent_state, use_call=False):
        cctxt = self.client.prepare(timeout=self.timeout)
        # add unique identifier to a report
        # that can be logged on server side.
        # This create visible correspondence between events on
        # the agent and on the server
        agent_state['uuid'] = uuidutils.generate_uuid()
        kwargs = {
            'agent_state': {'agent_state': agent_state},
            'time': datetime.utcnow().strftime(constants.ISO8601_TIME_FORMAT),
        }
        method = cctxt.call if use_call else cctxt.cast
        return method(context, 'report_state', **kwargs)


class PluginApi(object):
    '''Agent side of the rpc API.

    API version history:
        1.0 - Initial version.
        1.3 - get_device_details rpc signature upgrade to obtain 'host' and
              return value to include fixed_ips and device_owner for
              the device port
        1.4 - tunnel_sync rpc signature upgrade to obtain 'host'
        1.5 - Support update_device_list and
              get_devices_details_list_and_failed_devices
        1.6 - Support get_network_details
        1.7 - Support get_ports_by_vnic_type_and_host
        1.8 - Rename agent_restarted to refresh_tunnels in
              update_device_list to reflect its expanded purpose
        1.9 - Support for device definition as DeviceInfo(mac, pci_info) for:
              - get_device_details
              - get_devices_details_list (indirectly, calls get_device_details)
              - update_device_down
              - update_device_up
              - update_device_list (indirectly, called from update_device_down
                and update_device_up)
    '''

    def __init__(self, topic):
        target = oslo_messaging.Target(topic=topic, version='1.9')
        self.client = lib_rpc.get_client(target)

    def get_device_details(self, context, device, agent_id, host=None):
        cctxt = self.client.prepare(version='1.9')
        return cctxt.call(context, 'get_device_details', device=device,
                          agent_id=agent_id, host=host)

    def get_devices_details_list(self, context, devices, agent_id, host=None):
        cctxt = self.client.prepare(version='1.9')
        return cctxt.call(context, 'get_devices_details_list',
                          devices=devices, agent_id=agent_id, host=host)

    def get_devices_details_list_and_failed_devices(self, context, devices,
                                                    agent_id, host=None,
                                                    **kwargs):
        """Get devices details and the list of devices that failed.

        This method returns the devices details. If an error is thrown when
        retrieving the devices details, the device is put in a list of
        failed devices.
        """
        cctxt = self.client.prepare(version='1.5')
        return cctxt.call(
            context,
            'get_devices_details_list_and_failed_devices',
            devices=devices, agent_id=agent_id, host=host)

    def get_network_details(self, context, network, agent_id, host=None):
        cctxt = self.client.prepare(version='1.6')
        return cctxt.call(context, 'get_network_details', network=network,
                          agent_id=agent_id, host=host)

    def update_device_down(self, context, device, agent_id, host=None):
        cctxt = self.client.prepare(version='1.9')
        return cctxt.call(context, 'update_device_down', device=device,
                          agent_id=agent_id, host=host)

    def update_device_up(self, context, device, agent_id, host=None):
        cctxt = self.client.prepare(version='1.9')
        return cctxt.call(context, 'update_device_up', device=device,
                          agent_id=agent_id, host=host)

    def update_device_list(self, context, devices_up, devices_down,
                           agent_id, host, refresh_tunnels=False):
        cctxt = self.client.prepare(version='1.9')

        ret_devices_up = []
        failed_devices_up = []
        ret_devices_down = []
        failed_devices_down = []

        step = cfg.CONF.rpc_resources_processing_step
        devices_up = list(devices_up)
        devices_down = list(devices_down)
        for i in range(0, max(len(devices_up), len(devices_down)), step):
            # Divide-and-conquer RPC timeout
            ret = cctxt.call(context, 'update_device_list',
                             devices_up=devices_up[i:i + step],
                             devices_down=devices_down[i:i + step],
                             agent_id=agent_id, host=host,
                             refresh_tunnels=refresh_tunnels)
            ret_devices_up.extend(ret.get("devices_up", []))
            failed_devices_up.extend(ret.get("failed_devices_up", []))
            ret_devices_down.extend(ret.get("devices_down", []))
            failed_devices_down.extend(ret.get("failed_devices_down", []))

        return {'devices_up': ret_devices_up,
                'failed_devices_up': failed_devices_up,
                'devices_down': ret_devices_down,
                'failed_devices_down': failed_devices_down}

    def tunnel_sync(self, context, tunnel_ip, tunnel_type=None, host=None):
        cctxt = self.client.prepare(version='1.4')
        return cctxt.call(context, 'tunnel_sync', tunnel_ip=tunnel_ip,
                          tunnel_type=tunnel_type, host=host)

    def get_ports_by_vnic_type_and_host(self, context, vnic_type, host):
        cctxt = self.client.prepare(version='1.7')
        return cctxt.call(context, 'get_ports_by_vnic_type_and_host',
                          vnic_type=vnic_type, host=host)


class CacheBackedPluginApi(PluginApi):

    RESOURCE_TYPES = [resources.PORT,
                      resources.SECURITYGROUP,
                      resources.SECURITYGROUPRULE,
                      resources.NETWORK,
                      resources.SUBNET,
                      resources.ADDRESSGROUP]

    def __init__(self, *args, **kwargs):
        super(CacheBackedPluginApi, self).__init__(*args, **kwargs)
        self.remote_resource_cache = None
        self._create_cache_for_l2_agent()

    def register_legacy_notification_callbacks(self, legacy_interface):
        """Emulates the server-side notifications from ml2 AgentNotifierApi.

        legacy_interface is an object with 'delete'/'update' methods for
        core resources.
        """
        self._legacy_interface = legacy_interface
        for e in (callback_events.AFTER_UPDATE, callback_events.AFTER_DELETE):
            for r in (resources.PORT, resources.NETWORK):
                registry.subscribe(self._legacy_notifier, r, e)

    def _legacy_notifier(self, rtype, event, trigger, payload):
        """Checks if legacy interface is expecting calls for resource.

        looks for port_update, network_delete, etc and calls them with
        the payloads the handlers are expecting (an ID).
        """
        context = payload.context
        resource_id = payload.resource_id
        rtype = rtype.lower()  # all legacy handlers don't camelcase
        agent_restarted = payload.metadata.pop("agent_restarted", None)
        method, host_with_activation, host_with_deactivation = (
            self._get_method_host(rtype, event, payload))
        if not hasattr(self._legacy_interface, method):
            # TODO(kevinbenton): once these notifications are stable, emit
            # a deprecation warning for legacy handlers
            return
        # If there is a binding deactivation, we must also notify the
        # corresponding activation
        if method == BINDING_DEACTIVATE:
            self._legacy_interface.binding_deactivate(
                context, port_id=resource_id, host=host_with_deactivation)
            self._legacy_interface.binding_activate(
                context, port_id=resource_id, host=host_with_activation)
        else:
            payload = {rtype: {'id': resource_id},
                       '%s_id' % rtype: resource_id}
            if method == "port_update" and agent_restarted is not None:
                # Mark ovs-agent restart for local port_update
                payload["agent_restarted"] = agent_restarted
            getattr(self._legacy_interface, method)(context, **payload)

    def _get_method_host(self, rtype, event, payload):
        """Constructs the name of method to be called in the legacy interface.

        If the event received is a port update that contains a binding
        activation where a previous binding is deactivated, the method name
        is 'binding_deactivate' and the host where the binding has to be
        deactivated is returned. Otherwise, the method name is constructed from
        rtype and the event received and the host is None.
        """
        is_delete = event == callback_events.AFTER_DELETE
        suffix = 'delete' if is_delete else 'update'
        method = "%s_%s" % (rtype, suffix)
        host_with_activation = None
        host_with_deactivation = None
        if is_delete or rtype != callback_resources.PORT:
            return method, host_with_activation, host_with_deactivation

        # A port update was received. Find out if it is a binding activation
        # where a previous binding was deactivated
        BINDINGS = pb_ext.COLLECTION_NAME
        changed_fields = payload.metadata['changed_fields']
        if BINDINGS in changed_fields:
            existing_active_binding = (
                utils.get_port_binding_by_status_and_host(
                    getattr(payload.states[0], 'bindings', []),
                    constants.ACTIVE))
            updated_active_binding = (
                utils.get_port_binding_by_status_and_host(
                    getattr(payload.latest_state, 'bindings', []),
                    constants.ACTIVE))
            if (existing_active_binding and updated_active_binding and
                    existing_active_binding.host !=
                    updated_active_binding.host):
                if (utils.get_port_binding_by_status_and_host(
                        getattr(payload.latest_state, 'bindings', []),
                        constants.INACTIVE,
                        host=existing_active_binding.host)):
                    method = BINDING_DEACTIVATE
                    host_with_activation = updated_active_binding.host
                    host_with_deactivation = existing_active_binding.host
        return method, host_with_activation, host_with_deactivation

    def get_devices_details_list_and_failed_devices(self, context, devices,
                                                    agent_id, host=None,
                                                    agent_restarted=False):
        result = {'devices': [], 'failed_devices': []}
        for device in devices:
            try:
                result['devices'].append(
                    self.get_device_details(context, device, agent_id, host,
                                            agent_restarted))
            except Exception:
                LOG.exception("Failed to get details for device %s", device)
                result['failed_devices'].append(device)
        return result

    def get_device_details(self, context, device, agent_id, host=None,
                           agent_restarted=False):
        port_obj = self.remote_resource_cache.get_resource_by_id(
            resources.PORT, device, agent_restarted)
        if not port_obj:
            LOG.debug("Device %s does not exist in cache.", device)
            return {'device': device}
        if not port_obj.binding_levels:
            LOG.warning("Device %s is not bound.", port_obj)
            return {'device': device}
        segment = port_obj.binding_levels[-1].segment
        if not segment:
            LOG.debug("Device %s is not bound to any segment.", port_obj)
            return {'device': device}
        binding = utils.get_port_binding_by_status_and_host(
            port_obj.bindings, constants.ACTIVE, raise_if_not_found=True,
            port_id=port_obj.id)
        migrating_to = migrating_to_host(port_obj.bindings)
        if (not (migrating_to and cfg.CONF.nova.live_migration_events) and
                port_obj.device_owner.startswith(
                    constants.DEVICE_OWNER_COMPUTE_PREFIX) and
                binding[pb_ext.HOST] != host):
            LOG.debug("Device %s has no active binding in this host",
                      port_obj)
            return {'device': device,
                    constants.NO_ACTIVE_BINDING: True}
        net = self.remote_resource_cache.get_resource_by_id(
            resources.NETWORK, port_obj.network_id)
        net_qos_policy_id = net.qos_policy_id
        # match format of old RPC interface
        mac_addr = str(netaddr.EUI(str(port_obj.mac_address),
                                   dialect=netaddr.mac_unix_expanded))
        entry = {
            'device': device,
            'device_id': port_obj.device_id,
            'network_id': port_obj.network_id,
            'port_id': port_obj.id,
            'mac_address': mac_addr,
            'admin_state_up': port_obj.admin_state_up,
            'status': port_obj.status,
            'network_type': segment.network_type,
            'segmentation_id': segment.segmentation_id,
            'physical_network': segment.physical_network,
            'fixed_ips': [{'subnet_id': o.subnet_id,
                           'ip_address': str(o.ip_address)}
                          for o in port_obj.fixed_ips],
            'device_owner': port_obj.device_owner,
            'allowed_address_pairs': [{'mac_address': o.mac_address,
                                       'ip_address': o.ip_address}
                                      for o in port_obj.allowed_address_pairs],
            'port_security_enabled': getattr(port_obj.security,
                                             'port_security_enabled', True),
            'qos_policy_id': port_obj.qos_policy_id,
            'network_qos_policy_id': net_qos_policy_id,
            'profile': binding.profile,
            'vif_type': binding.vif_type,
            'vnic_type': binding.vnic_type,
            'security_groups': list(port_obj.security_group_ids),
            'migrating_to': migrating_to,
        }
        LOG.debug("Returning: %s", entry)
        return entry

    def get_devices_details_list(self, context, devices, agent_id, host=None):
        return [self.get_device_details(context, device, agent_id, host)
                for device in devices]

    def _create_cache_for_l2_agent(self):
        """Create a push-notifications cache for L2 agent related resources."""
        objects.register_objects()
        rcache = resource_cache.RemoteResourceCache(self.RESOURCE_TYPES)
        rcache.start_watcher()
        self.remote_resource_cache = rcache


# TODO(ralonsoh): move this method to neutron_lib.plugins.utils
def migrating_to_host(bindings, host=None):
    """Return the host the port is being migrated.

    If the host is passed, the port binding profile with the "migrating_to",
    that contains the host the port is being migrated, is compared to this
    value. If no value is passed, this method will return if the port is
    being migrated ("migrating_to" is present in any port binding profile).

    The function returns None or the matching host.
    """
    for binding in (binding for binding in bindings if
                    binding[pb_ext.STATUS] == constants.ACTIVE):
        profile = binding.get('profile')
        if not profile:
            continue
        profile = (jsonutils.loads(profile) if isinstance(profile, str) else
                   profile)
        migrating_to = profile.get('migrating_to')
        if migrating_to:
            if not host:  # Just know if the port is being migrated.
                return migrating_to
            if migrating_to == host:
                return migrating_to
    return None



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_config import cfg

from neutron._i18n import _


L3_AGENTS_SCHEDULER_OPTS = [
    cfg.StrOpt('router_scheduler_driver',
               default='neutron.scheduler.l3_agent_scheduler.'
                       'LeastRoutersScheduler',
               help=_('Driver to use for scheduling '
                      'router to a default L3 agent')),
    cfg.BoolOpt('router_auto_schedule', default=True,
                help=_('Allow auto scheduling of routers to L3 agent.')),
    cfg.BoolOpt('allow_automatic_l3agent_failover', default=False,
                help=_('Automatically reschedule routers from offline L3 '
                       'agents to online L3 agents.')),
]


def register_db_l3agentschedulers_opts(conf=cfg.CONF):
    conf.register_opts(L3_AGENTS_SCHEDULER_OPTS)



# Copyright (C) 2013 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import contextlib

from neutron_lib.api.definitions import metering as metering_apidef
from neutron_lib import constants as n_consts
from neutron_lib import context
from neutron_lib.db import constants as db_const
from neutron_lib.plugins import constants
from oslo_utils import uuidutils
import webob.exc

from neutron.api import extensions
from neutron.common import config
import neutron.extensions
from neutron.services.metering import metering_plugin
from neutron.tests.unit.db import test_db_base_plugin_v2

DB_METERING_PLUGIN_KLASS = (
    "neutron.services.metering."
    "metering_plugin.MeteringPlugin"
)

extensions_path = ':'.join(neutron.extensions.__path__)
_long_description_ok = 'x' * (db_const.LONG_DESCRIPTION_FIELD_SIZE)
_long_description_ng = 'x' * (db_const.LONG_DESCRIPTION_FIELD_SIZE + 1)
_fake_uuid = uuidutils.generate_uuid


class MeteringPluginDbTestCaseMixin(object):
    def _create_metering_label(self, fmt, name, description, **kwargs):
        data = {'metering_label': {'name': name,
                                   'tenant_id': kwargs.get('tenant_id',
                                                           'test-tenant'),
                                   'shared': kwargs.get('shared', False),
                                   'description': description}}
        req = self.new_create_request('metering-labels', data,
                                      fmt)

        if kwargs.get('set_context') and 'tenant_id' in kwargs:
            # create a specific auth context for this request
            req.environ['neutron.context'] = (
                context.Context('', kwargs['tenant_id'],
                                is_admin=kwargs.get('is_admin', True)))

        return req.get_response(self.ext_api)

    def _make_metering_label(self, fmt, name, description, **kwargs):
        res = self._create_metering_label(fmt, name, description, **kwargs)
        if res.status_int >= 400:
            raise webob.exc.HTTPClientError(code=res.status_int)
        return self.deserialize(fmt, res)

    def _create_metering_label_rule(self, fmt, metering_label_id, direction,
                                    excluded, remote_ip_prefix=None,
                                    source_ip_prefix=None,
                                    destination_ip_prefix=None,
                                    **kwargs):
        data = {
            'metering_label_rule': {
                        'metering_label_id': metering_label_id,
                        'tenant_id': kwargs.get('tenant_id', 'test-tenant'),
                        'direction': direction,
                        'excluded': excluded,
                     }
                }

        if remote_ip_prefix:
            data['metering_label_rule']['remote_ip_prefix'] = remote_ip_prefix

        if source_ip_prefix:
            data['metering_label_rule']['source_ip_prefix'] = source_ip_prefix

        if destination_ip_prefix:
            data['metering_label_rule']['destination_ip_prefix'] =\
                destination_ip_prefix

        req = self.new_create_request('metering-label-rules',
                                      data, fmt)

        if kwargs.get('set_context') and 'tenant_id' in kwargs:
            # create a specific auth context for this request
            req.environ['neutron.context'] = (
                context.Context('', kwargs['tenant_id']))

        return req.get_response(self.ext_api)

    def _make_metering_label_rule(self, fmt, metering_label_id, direction,
                                  excluded, **kwargs):
        res = self._create_metering_label_rule(fmt, metering_label_id,
                                               direction, excluded, **kwargs)
        if res.status_int >= 400:
            raise webob.exc.HTTPClientError(code=res.status_int)
        return self.deserialize(fmt, res)

    @contextlib.contextmanager
    def metering_label(self, name='label', description='desc',
                       fmt=None, **kwargs):
        if not fmt:
            fmt = self.fmt
        metering_label = self._make_metering_label(fmt, name,
                                                   description, **kwargs)
        yield metering_label

    @contextlib.contextmanager
    def metering_label_rule(self, metering_label_id=None, direction='ingress',
                            excluded=False, fmt=None, **kwargs):
        if not fmt:
            fmt = self.fmt
        metering_label_rule = self._make_metering_label_rule(fmt,
                                                             metering_label_id,
                                                             direction,
                                                             excluded,
                                                             **kwargs)
        yield metering_label_rule


class MeteringPluginDbTestCase(
        test_db_base_plugin_v2.NeutronDbPluginV2TestCase,
        MeteringPluginDbTestCaseMixin):
    fmt = 'json'

    resource_prefix_map = dict(
        (k.replace('_', '-'), "/metering")
        for k in metering_apidef.RESOURCE_ATTRIBUTE_MAP.keys()
    )

    def setUp(self, plugin=None):
        service_plugins = {'metering_plugin_name': DB_METERING_PLUGIN_KLASS}

        super(MeteringPluginDbTestCase, self).setUp(
            plugin=plugin,
            service_plugins=service_plugins
        )

        self.plugin = metering_plugin.MeteringPlugin()
        ext_mgr = extensions.PluginAwareExtensionManager(
            extensions_path,
            {constants.METERING: self.plugin}
        )
        app = config.load_paste_app('extensions_test_app')
        self.ext_api = extensions.ExtensionMiddleware(app, ext_mgr=ext_mgr)


class TestMetering(MeteringPluginDbTestCase):
    def test_create_metering_label(self):
        name = 'my label'
        description = 'my metering label'
        keys = [('name', name,), ('description', description)]
        with self.metering_label(name, description) as metering_label:
            for k, v, in keys:
                self.assertEqual(metering_label['metering_label'][k], v)

    def test_create_metering_label_shared(self):
        name = 'my label'
        description = 'my metering label'
        shared = True
        keys = [('name', name,), ('description', description),
                ('shared', shared)]
        with self.metering_label(name, description,
                                 shared=shared) as metering_label:
            for k, v, in keys:
                self.assertEqual(metering_label['metering_label'][k], v)

    def test_create_metering_label_with_max_description_length(self):
        res = self._create_metering_label(self.fmt, 'my label',
                                          _long_description_ok)
        self.assertEqual(webob.exc.HTTPCreated.code, res.status_int)

    def test_create_metering_label_with_too_long_description(self):
        res = self._create_metering_label(self.fmt, 'my label',
                                          _long_description_ng)
        self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_update_metering_label(self):
        name = 'my label'
        description = 'my metering label'
        data = {'metering_label': {}}
        with self.metering_label(name, description) as metering_label:
            metering_label_id = metering_label['metering_label']['id']
            self._update('metering-labels', metering_label_id, data,
                         webob.exc.HTTPNotImplemented.code)

    def test_delete_metering_label(self):
        name = 'my label'
        description = 'my metering label'

        with self.metering_label(name, description) as metering_label:
            metering_label_id = metering_label['metering_label']['id']
            self._delete('metering-labels', metering_label_id, 204)

    def test_list_metering_label(self):
        name = 'my label'
        description = 'my metering label'

        with self.metering_label(name, description) as v1,\
                self.metering_label(name, description) as v2:
            metering_label = (v1, v2)

            self._test_list_resources('metering-label', metering_label)

    def test_create_metering_label_rule(self):
        name = 'my label'
        description = 'my metering label'

        with self.metering_label(name, description) as metering_label:
            metering_label_id = metering_label['metering_label']['id']

            direction = 'egress'
            remote_ip_prefix = '192.168.0.0/24'
            excluded = True

            keys = [('metering_label_id', metering_label_id),
                    ('direction', direction),
                    ('excluded', excluded),
                    ('remote_ip_prefix', remote_ip_prefix)]
            with self.metering_label_rule(
                    metering_label_id, direction, excluded,
                    remote_ip_prefix=remote_ip_prefix) as label_rule:
                for k, v, in keys:
                    self.assertEqual(label_rule['metering_label_rule'][k], v)

    def test_create_metering_label_rule_with_non_existent_label(self):
        direction = 'egress'
        remote_ip_prefix = '192.168.0.0/24'
        excluded = True

        res = self._create_metering_label_rule(
            self.fmt, _fake_uuid(), direction, excluded,
            remote_ip_prefix=remote_ip_prefix)
        self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)

    def test_update_metering_label_rule(self):
        name = 'my label'
        description = 'my metering label'
        direction = 'egress'
        remote_ip_prefix = '192.168.0.0/24'
        data = {'metering_label_rule': {}}
        with self.metering_label(name, description) as metering_label, \
                self.metering_label_rule(
                    metering_label['metering_label']['id'], direction,
                    remote_ip_prefix=remote_ip_prefix) as label_rule:
            rule_id = label_rule['metering_label_rule']['id']
            self._update('metering-label-rules', rule_id, data,
                         webob.exc.HTTPNotImplemented.code)

    def test_delete_metering_label_rule(self):
        name = 'my label'
        description = 'my metering label'

        with self.metering_label(name, description) as metering_label:
            metering_label_id = metering_label['metering_label']['id']

            direction = 'egress'
            remote_ip_prefix = '192.168.0.0/24'
            excluded = True

            with self.metering_label_rule(
                    metering_label_id, direction, excluded,
                    remote_ip_prefix=remote_ip_prefix) as label_rule:
                rule_id = label_rule['metering_label_rule']['id']
                self._delete('metering-label-rules', rule_id, 204)

    def test_list_metering_label_rule(self):
        name = 'my label'
        description = 'my metering label'

        with self.metering_label(name, description) as metering_label:
            metering_label_id = metering_label['metering_label']['id']

            direction = 'egress'
            remote_ip_prefix = '192.168.0.0/24'
            excluded = True

            with self.metering_label_rule(
                    metering_label_id, direction, excluded,
                    remote_ip_prefix=remote_ip_prefix) as v1,\
                    self.metering_label_rule(
                        metering_label_id, 'ingress', excluded,
                        remote_ip_prefix=remote_ip_prefix) as v2:
                metering_label_rule = (v1, v2)

                self._test_list_resources('metering-label-rule',
                                          metering_label_rule)

    def test_create_metering_label_rules(self):
        name = 'my label'
        description = 'my metering label'

        with self.metering_label(name, description) as metering_label:
            metering_label_id = metering_label['metering_label']['id']

            direction = 'egress'
            remote_ip_prefix = '192.168.0.0/24'
            excluded = True

            with self.metering_label_rule(
                    metering_label_id, direction, excluded,
                    remote_ip_prefix=remote_ip_prefix) as v1,\
                    self.metering_label_rule(
                        metering_label_id, direction, False,
                        remote_ip_prefix=n_consts.IPv4_ANY) as v2:
                metering_label_rule = (v1, v2)

                self._test_list_resources('metering-label-rule',
                                          metering_label_rule)

    def test_create_overlap_metering_label_rules(self):
        name = 'my label'
        description = 'my metering label'

        with self.metering_label(name, description) as metering_label:
            metering_label_id = metering_label['metering_label']['id']

            direction = 'egress'
            remote_ip_prefix1 = '192.168.0.0/24'
            remote_ip_prefix2 = '192.168.0.0/16'
            excluded = True

            with self.metering_label_rule(
                    metering_label_id, direction, excluded,
                    remote_ip_prefix=remote_ip_prefix1):
                res = self._create_metering_label_rule(
                    self.fmt, metering_label_id, direction, excluded,
                    remote_ip_prefix=remote_ip_prefix2)
                self.assertEqual(webob.exc.HTTPConflict.code, res.status_int)

    def test_create_metering_label_rule_two_labels(self):
        name1 = 'my label 1'
        name2 = 'my label 2'
        description = 'my metering label'

        with self.metering_label(name1, description) as metering_label1:
            metering_label_id1 = metering_label1['metering_label']['id']

            with self.metering_label(name2, description) as metering_label2:
                metering_label_id2 = metering_label2['metering_label']['id']

                direction = 'egress'
                remote_ip_prefix = '192.168.0.0/24'
                excluded = True

                with self.metering_label_rule(
                        metering_label_id1, direction, excluded,
                        remote_ip_prefix=remote_ip_prefix) as v1,\
                        self.metering_label_rule(
                            metering_label_id2, direction, excluded,
                            remote_ip_prefix=remote_ip_prefix) as v2:
                    metering_label_rule = (v1, v2)

                    self._test_list_resources('metering-label-rule',
                                              metering_label_rule)



# Copyright 2020 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
import sqlalchemy as sa


"""propagate_uplink_status_defaults_to_true

Revision ID: 49d8622c5221
Revises: 1ea5dab0897a
Create Date: 2020-07-31 15:27:29.425953

"""

# revision identifiers, used by Alembic.
revision = '49d8622c5221'
down_revision = '1ea5dab0897a'


def upgrade():
    op.alter_column('portuplinkstatuspropagation', 'propagate_uplink_status',
                    existing_type=sa.Boolean(), existing_nullable=False,
                    server_default=sa.sql.true())



# Copyright (c) 2014 OpenStack Foundation, all rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from neutron_lib.api.definitions import portbindings
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as const
from neutron_lib import context
from neutron_lib.db import api as db_api
from neutron_lib import exceptions
from neutron_lib.exceptions import l3 as l3_exc
from neutron_lib.objects import exceptions as o_exc
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from neutron_lib.plugins import utils as plugin_utils
from oslo_utils import uuidutils

from neutron.db import agents_db
from neutron.db import l3_dvr_db
from neutron.db import l3_dvrscheduler_db
from neutron.db.models import l3 as l3_models
from neutron.db import models_v2
from neutron.objects import agent as agent_obj
from neutron.objects import l3agent as rb_obj
from neutron.objects import ports as port_obj
from neutron.objects import router as router_obj
from neutron.tests.unit.db import test_db_base_plugin_v2
from neutron.tests.unit.extensions import test_l3

_uuid = uuidutils.generate_uuid


class FakeL3Plugin(test_l3.TestL3PluginBaseAttributes,
                   l3_dvr_db.L3_NAT_with_dvr_db_mixin,
                   l3_dvrscheduler_db.L3_DVRsch_db_mixin,
                   agents_db.AgentDbMixin):
    pass


class L3DvrTestCase(test_db_base_plugin_v2.NeutronDbPluginV2TestCase):

    def setUp(self):
        super(L3DvrTestCase, self).setUp(plugin='ml2')
        self.core_plugin = directory.get_plugin()
        self.ctx = context.get_admin_context()
        self.mixin = FakeL3Plugin()
        directory.add_plugin(plugin_constants.L3, self.mixin)

    def _create_router(self, router):
        with db_api.CONTEXT_WRITER.using(self.ctx):
            return self.mixin._create_router_db(self.ctx, router, 'foo_tenant')

    def create_port(self, net_id, port_info):
        with db_api.CONTEXT_WRITER.using(self.ctx):
            return self._create_port(self.fmt, net_id, **port_info)

    def _test__create_router_db(self, expected=False, distributed=None):
        router = {'name': 'foo_router', 'admin_state_up': True}
        if distributed is not None:
            router['distributed'] = distributed
        result = self._create_router(router)
        self.assertEqual(expected, result.extra_attributes['distributed'])

    def test_create_router_db_default(self):
        self._test__create_router_db(expected=False)

    def test_create_router_db_centralized(self):
        self._test__create_router_db(expected=False, distributed=False)

    def test_create_router_db_distributed(self):
        self._test__create_router_db(expected=True, distributed=True)

    def _test__validate_router_migration_on_router_update(self, mock_arg):
        router = {
            'name': 'foo_router',
            'admin_state_up': True,
            'distributed': True
        }
        router_db = self._create_router(router)
        self.assertFalse(self.mixin._validate_router_migration(
            self.ctx, router_db, {'name': 'foo_router_2'}))

    # mock the check function to indicate that the variable
    # _admin_state_down_necessary set to True
    @mock.patch('neutron.db.l3_dvr_db.is_admin_state_down_necessary',
                return_value=True)
    def test__validate_router_migration_on_router_update_mock(self,
            mock_arg):
        # call test with admin_state_down_before_update ENABLED
        self._test__validate_router_migration_on_router_update(mock_arg)

    # mock the check function to indicate that the variable
    # _admin_state_down_necessary set to False
    @mock.patch('neutron.db.l3_dvr_db.is_admin_state_down_necessary',
                return_value=False)
    def test__validate_router_migration_on_router_update(self, mock_arg):
        # call test with admin_state_down_before_update DISABLED
        self._test__validate_router_migration_on_router_update(mock_arg)

    def _test__validate_router_migration_raise_error(self):
        router = {
            'name': 'foo_router',
            'admin_state_up': True,
            'distributed': True
        }
        router_db = self._create_router(router)
        self.assertRaises(exceptions.BadRequest,
                          self.mixin._validate_router_migration,
                          self.ctx, router_db, {'distributed': False})

    @mock.patch('neutron.db.l3_dvr_db.is_admin_state_down_necessary',
                return_value=True)
    def test__validate_router_migration_raise_error_mocked(self, mock_arg):
        # call test with admin_state_down_before_update ENABLED
        self._test__validate_router_migration_raise_error()

    @mock.patch('neutron.db.l3_dvr_db.is_admin_state_down_necessary',
                return_value=False)
    def test__validate_router_migration_raise_error(self, mock_arg):
        # call test with admin_state_down_before_update DISABLED
        self._test__validate_router_migration_raise_error()

    @mock.patch('neutron.db.l3_dvr_db.is_admin_state_down_necessary',
                return_value=True)
    def test__validate_router_migration_old_router_up_raise_error(self,
            mock_arg):
        # call test with admin_state_down_before_update ENABLED
        old_router = {
            'name': 'bar_router',
            'admin_state_up': True,
            'distributed': True
        }
        new_router = {
            'name': 'foo_router',
            'admin_state_up': False,
            'distributed': False
        }
        update = {'distributed': False}
        router_db = self._create_router(new_router)
        self.assertRaises(exceptions.BadRequest,
                          self.mixin._validate_router_migration,
                          self.ctx, router_db, update,
                          old_router)

    def _test_upgrade_inactive_router_to_distributed_validation_success(self):
        router = {'name': 'foo_router', 'admin_state_up': False,
                 'distributed': False}
        router_db = self._create_router(router)
        update = {'distributed': True}
        self.assertTrue(self.mixin._validate_router_migration(
            self.ctx, router_db, update))

    @mock.patch('neutron.db.l3_dvr_db.is_admin_state_down_necessary',
                return_value=True)
    def test_upgrade_inactive_router_to_distributed_validation_success_mocked(
            self, mock_arg):
        # call test with admin_state_down_before_update ENABLED
        self._test_upgrade_inactive_router_to_distributed_validation_success()

    @mock.patch('neutron.db.l3_dvr_db.is_admin_state_down_necessary',
                return_value=False)
    def test_upgrade_inactive_router_to_distributed_validation_success(self,
            mock_arg):
        # call test with admin_state_down_before_update DISABLED
        self._test_upgrade_inactive_router_to_distributed_validation_success()

    def _test_upgrade_active_router_to_distributed_validation_failure(self):
        router = {'name': 'foo_router', 'admin_state_up': True,
                 'distributed': False}
        router_db = self._create_router(router)
        update = {'distributed': True}
        self.assertRaises(exceptions.BadRequest,
                          self.mixin._validate_router_migration,
                          self.ctx, router_db, update)

    @mock.patch('neutron.db.l3_dvr_db.is_admin_state_down_necessary',
                return_value=True)
    def test_upgrade_active_router_to_distributed_validation_failure(self,
            mock_arg):
        # call test with admin_state_down_before_update ENABLED
        self._test_upgrade_active_router_to_distributed_validation_failure()

    @mock.patch('neutron.db.l3_dvr_db.is_admin_state_down_necessary',
                return_value=True)
    def test_downgrade_active_router_to_centralized_validation_failure(self,
            mock_arg):
        # call test with admin_state_down_before_update ENABLED
        router = {'name': 'foo_router', 'admin_state_up': True,
                'distributed': True}
        router_db = self._create_router(router)
        update = {'distributed': False}
        self.assertRaises(exceptions.BadRequest,
                          self.mixin._validate_router_migration,
                          self.ctx, router_db, update)

    def test_update_router_db_centralized_to_distributed(self):
        router = {'name': 'foo_router', 'admin_state_up': True}
        agent = {'id': _uuid()}
        distributed = {'distributed': True}
        router_db = self._create_router(router)
        router_id = router_db['id']
        self.assertFalse(router_db.extra_attributes.distributed)
        self.mixin._get_router = mock.Mock(return_value=router_db)
        self.mixin._validate_router_migration = mock.Mock()
        self.mixin._migrate_router_ports = mock.Mock()
        self.mixin.list_l3_agents_hosting_router = mock.Mock(
            return_value={'agents': [agent]})
        self.mixin._unbind_router = mock.Mock()
        router_db = self.mixin._update_router_db(
            self.ctx, router_id, distributed)
        # Assert that the DB value has changed
        self.assertTrue(router_db.extra_attributes.distributed)
        self.assertEqual(1,
                         self.mixin._migrate_router_ports.call_count)

    def test_update_router_db_distributed_to_centralized(self):
        router = {'name': 'foo_router', 'admin_state_up': True,
                  'distributed': True}
        agent = {'id': _uuid(), 'host': 'xyz'}
        router_db = self._create_router(router)
        router_id = router_db['id']
        self.assertTrue(router_db.extra_attributes.distributed)
        self.mixin._get_router = mock.Mock(return_value=router_db)
        self.mixin._validate_router_migration = mock.Mock()
        self.mixin._migrate_router_ports = mock.Mock()
        self.mixin._core_plugin.\
            delete_distributed_port_bindings_by_router_id = mock.Mock()
        self.mixin.list_l3_agents_hosting_router = mock.Mock(
            return_value={'agents': [agent]})
        self.mixin._unbind_router = mock.Mock()
        updated_router = self.mixin.update_router(self.ctx, router_id,
            {'router': {'distributed': False}})
        # Assert that the DB value has changed
        self.assertFalse(updated_router['distributed'])
        self.assertEqual(1,
                         self.mixin._migrate_router_ports.call_count)
        self.assertEqual(
            1,
            self.mixin._core_plugin.
            delete_distributed_port_bindings_by_router_id.call_count)

    def _test_get_device_owner(self, is_distributed=False,
                               expected=const.DEVICE_OWNER_ROUTER_INTF,
                               pass_router_id=True):
        router = {
            'name': 'foo_router',
            'admin_state_up': True,
            'distributed': is_distributed
        }
        router_db = self._create_router(router)
        router_pass = router_db['id'] if pass_router_id else router_db
        with mock.patch.object(self.mixin, '_get_router') as f:
            f.return_value = router_db
            result = self.mixin._get_device_owner(self.ctx, router_pass)
            self.assertEqual(expected, result)

    def test_get_device_owner_by_router_id(self):
        self._test_get_device_owner()

    def test__get_device_owner_centralized(self):
        self._test_get_device_owner(pass_router_id=False)

    def test__get_device_owner_distributed(self):
        self._test_get_device_owner(
            is_distributed=True,
            expected=const.DEVICE_OWNER_DVR_INTERFACE,
            pass_router_id=False)

    def _test__is_distributed_router(self, router, expected):
        result = l3_dvr_db.is_distributed_router(router)
        self.assertEqual(expected, result)

    def test__is_distributed_router_by_db_object(self):
        router = {'name': 'foo_router', 'admin_state_up': True}
        router_db = self._create_router(router)
        self.mixin._get_device_owner(mock.ANY, router_db)

    def test__is_distributed_router_default(self):
        router = {'id': 'foo_router_id'}
        self._test__is_distributed_router(router, False)

    def test__is_distributed_router_centralized(self):
        router = {'id': 'foo_router_id', 'distributed': False}
        self._test__is_distributed_router(router, False)

    def test__is_distributed_router_distributed(self):
        router = {'id': 'foo_router_id', 'distributed': True}
        self._test__is_distributed_router(router, True)

    def test__get_agent_gw_ports_exist_for_network(self):
        plugin = mock.Mock()
        directory.add_plugin(plugin_constants.CORE, plugin)
        plugin.get_ports.return_value = []
        self.mixin._get_agent_gw_ports_exist_for_network(
            self.ctx, 'network_id', 'host', 'agent_id')
        plugin.get_ports.assert_called_with(self.ctx, {
            'network_id': ['network_id'],
            'device_id': ['agent_id'],
            'device_owner': [const.DEVICE_OWNER_AGENT_GW]})

    def _help_check_and_create_fip_gw_port(self, fip=None):
        port = {
            'id': '1234',
            portbindings.HOST_ID: 'myhost',
            'floating_network_id': 'external_net'
        }
        ctxt = mock.Mock()
        with mock.patch.object(self.mixin,
                'create_fip_agent_gw_port_if_not_exists') as c_fip,\
                mock.patch.object(router_obj.FloatingIP, 'get_objects',
                                  return_value=[fip] if fip else None):
            (self.mixin.
             check_for_fip_and_create_agent_gw_port_on_host_if_not_exists(
                    ctxt, port, 'host'))
            if fip:
                c_fip.assert_called_once_with(ctxt.elevated(),
                                              fip['floating_network_id'],
                                              'host')
            else:
                c_fip.assert_not_called()

    def test_check_for_fip_and_create_agent_gw_port_no_fip(self):
        self._help_check_and_create_fip_gw_port()

    def test_check_for_fip_and_create_agent_gw_port_with_dvr_true(self):
        fip = {
            'id': _uuid(),
            'floating_network_id': 'fake_net_id',
            'router_id': 'foo_router_id'
        }
        self._help_check_and_create_fip_gw_port(fip=fip)

    def _test_prepare_direct_delete_dvr_internal_ports(self, port):
        plugin = mock.Mock()
        directory.add_plugin(plugin_constants.CORE, plugin)
        plugin.get_port.return_value = port
        with mock.patch.object(router_obj.Router, 'objects_exist',
                               return_value=True):
            self.assertRaises(exceptions.ServicePortInUse,
                              self.mixin.prevent_l3_port_deletion,
                              self.ctx, port['id'])

    def test_prevent_delete_floatingip_agent_gateway_port(self):
        port = {
            'id': 'my_port_id',
            'fixed_ips': mock.ANY,
            'device_id': 'r_id',
            'device_owner': const.DEVICE_OWNER_AGENT_GW
        }
        self._test_prepare_direct_delete_dvr_internal_ports(port)

    def test_prevent_delete_csnat_port(self):
        port = {
            'id': 'my_port_id',
            'fixed_ips': mock.ANY,
            'device_id': 'r_id',
            'device_owner': const.DEVICE_OWNER_ROUTER_SNAT
        }
        self._test_prepare_direct_delete_dvr_internal_ports(port)

    def test__create_gw_port_with_no_gateway(self):
        router = {
            'name': 'foo_router',
            'admin_state_up': True,
            'distributed': True,
        }
        router_db = self._create_router(router)
        router_id = router_db['id']
        self.assertTrue(router_db.extra_attributes.distributed)
        with mock.patch.object(l3_dvr_db.l3_db.L3_NAT_db_mixin,
                               '_create_gw_port'),\
                mock.patch.object(
                    self.mixin,
                    '_create_snat_intf_ports_if_not_exists') as cs:
            self.mixin._create_gw_port(
                self.ctx, router_id, router_db, mock.ANY,
                mock.ANY, mock.ANY)
            self.assertFalse(cs.call_count)

    def test_build_routers_list_with_gw_port_mismatch(self):
        routers = [{'gw_port_id': 'foo_gw_port_id', 'id': 'foo_router_id'}]
        gw_ports = {}
        routers = self.mixin._build_routers_list(self.ctx, routers, gw_ports)
        self.assertIsNone(routers[0].get('gw_port'))

    def _helper_delete_floatingip_agent_gateway_port(
            self, port_host, delete_dvr_fip_agent_port_side_effect=None):
        ports = [{
            'id': 'my_port_id',
            portbindings.HOST_ID: 'foo_host',
            'network_id': 'ext_network_id',
            'device_owner': const.DEVICE_OWNER_ROUTER_GW
        },
                {
            'id': 'my_new_port_id',
            portbindings.HOST_ID: 'my_foo_host',
            'network_id': 'ext_network_id',
            'device_owner': const.DEVICE_OWNER_ROUTER_GW
        }]
        plugin = mock.Mock()
        directory.add_plugin(plugin_constants.CORE, plugin)
        plugin.get_ports.return_value = ports
        self.mixin._get_agent_by_type_and_host = mock.Mock(
            return_value={'id': uuidutils.generate_uuid()})

        with mock.patch.object(
            router_obj, "DvrFipGatewayPortAgentBinding"
        ) as dvr_fip_agent_port_obj:
            dvr_fip_agent_port_obj_instance = (
                dvr_fip_agent_port_obj.return_value)
            dvr_fip_agent_port_obj_instance.delete.side_effect = (
                delete_dvr_fip_agent_port_side_effect)

            self.mixin.delete_floatingip_agent_gateway_port(
                self.ctx, port_host, 'ext_network_id')

        plugin.get_ports.assert_called_with(self.ctx, filters={
            'network_id': ['ext_network_id'],
            'device_owner': [const.DEVICE_OWNER_AGENT_GW]})
        if port_host:
            plugin.ipam.delete_port.assert_called_once_with(
                self.ctx, 'my_port_id')
            dvr_fip_agent_port_obj_instance.delete.assert_called_once()
        else:
            plugin.ipam.delete_port.assert_called_with(
                self.ctx, 'my_new_port_id')
            dvr_fip_agent_port_obj_instance.delete.assert_called()

    def test_delete_floatingip_agent_gateway_port_without_host_id(self):
        self._helper_delete_floatingip_agent_gateway_port(None)

    def test_delete_floatingip_agent_gateway_port_with_host_id(self):
        self._helper_delete_floatingip_agent_gateway_port(
            'foo_host')

    def test_delete_floatingip_agent_gateway_port_no_host_id_fip_gw_not_found(
            self):
        self._helper_delete_floatingip_agent_gateway_port(
                None, exceptions.ObjectNotFound(id='my_port_id'))

    def test_delete_floatingip_agent_gateway_port_host_id_fip_gw_not_found(
            self):
        self._helper_delete_floatingip_agent_gateway_port(
            'foo_host', exceptions.ObjectNotFound(id='my_port_id'))

    def _setup_delete_current_gw_port_deletes_dvr_internal_ports(
            self, port=None, gw_port=True, new_network_id='ext_net_id_2'):
        router_db = {
            'name': 'foo_router',
            'admin_state_up': True,
            'distributed': True
        }
        with db_api.CONTEXT_WRITER.using(self.ctx):
            router = self._create_router(router_db)
            if gw_port:
                with self.subnet(cidr='10.10.10.0/24') as subnet:
                    port_dict = {
                        'device_id': router.id,
                        'device_owner': const.DEVICE_OWNER_ROUTER_GW,
                        'admin_state_up': True,
                        'fixed_ips': [{'subnet_id': subnet['subnet']['id'],
                                       'ip_address': '10.10.10.100'}]
                    }
                net_id = subnet['subnet']['network_id']
                port_res = self.create_port(net_id, port_dict)
                port_res_dict = self.deserialize(self.fmt, port_res)
                port_db = self.ctx.session.query(models_v2.Port).filter_by(
                    id=port_res_dict['port']['id']).one()
                router.gw_port = port_db
                router_port = l3_models.RouterPort(
                    router_id=router.id,
                    port_id=port_db.id,
                    port_type=const.DEVICE_OWNER_ROUTER_GW
                )
                self.ctx.session.add(router)
                self.ctx.session.add(router_port)
            else:
                net_id = None

        plugin = mock.Mock()
        directory.add_plugin(plugin_constants.CORE, plugin)
        with mock.patch.object(l3_dvr_db.l3_db.L3_NAT_db_mixin,
                               'router_gw_port_has_floating_ips',
                               return_value=False),\
            mock.patch.object(
                self.mixin,
                '_get_router') as grtr,\
            mock.patch.object(
                self.mixin,
                'delete_csnat_router_interface_ports') as del_csnat_port,\
            mock.patch.object(
                self.mixin,
                'delete_floatingip_agent_gateway_port') as del_agent_gw_port,\
            mock.patch.object(
                self.mixin.l3_rpc_notifier,
                'delete_fipnamespace_for_ext_net') as del_fip:
            plugin.get_ports.return_value = port
            grtr.return_value = router
            self.mixin._delete_current_gw_port(
                self.ctx, router['id'], router, new_network_id)
            return router, plugin, net_id, del_csnat_port,\
                del_agent_gw_port, del_fip

    def test_delete_current_gw_port_deletes_fip_agent_gw_port_and_fipnamespace(
            self):
        rtr, plugin, ext_net_id, d_csnat_port, d_agent_gw_port, del_fip = (
            self._setup_delete_current_gw_port_deletes_dvr_internal_ports())
        self.assertFalse(d_csnat_port.called)
        self.assertTrue(d_agent_gw_port.called)
        d_agent_gw_port.assert_called_once_with(mock.ANY, None, ext_net_id)
        del_fip.assert_called_once_with(self.ctx, ext_net_id)

    def test_delete_current_gw_port_never_calls_delete_fip_agent_gw_port(self):
        port = [{
            'id': 'my_port_id',
            'network_id': 'ext_net_id',
            'device_owner': const.DEVICE_OWNER_ROUTER_GW
        },
                {
            'id': 'my_new_port_id',
            'network_id': 'ext_net_id',
            'device_owner': const.DEVICE_OWNER_ROUTER_GW
        }]
        rtr, plugin, ext_net_id, d_csnat_port, d_agent_gw_port, del_fip = (
            self._setup_delete_current_gw_port_deletes_dvr_internal_ports(
                port=port))
        self.assertFalse(d_csnat_port.called)
        self.assertFalse(d_agent_gw_port.called)
        self.assertFalse(del_fip.called)
        self.assertIsNotNone(ext_net_id)

    def test_delete_current_gw_port_never_calls_delete_fipnamespace(self):
        rtr, plugin, ext_net_id, d_csnat_port, d_agent_gw_port, del_fip = (
            self._setup_delete_current_gw_port_deletes_dvr_internal_ports(
                gw_port=False))
        self.assertFalse(d_csnat_port.called)
        self.assertFalse(d_agent_gw_port.called)
        self.assertFalse(del_fip.called)
        self.assertIsNone(ext_net_id)

    def test_delete_current_gw_port_deletes_csnat_port(self):
        rtr, plugin, ext_net_id, d_csnat_port, d_agent_gw_port, del_fip = (
            self._setup_delete_current_gw_port_deletes_dvr_internal_ports(
                new_network_id=None))
        self.assertTrue(d_csnat_port.called)
        self.assertTrue(d_agent_gw_port.called)
        d_csnat_port.assert_called_once_with(mock.ANY, rtr)
        d_agent_gw_port.assert_called_once_with(mock.ANY, None, ext_net_id)
        del_fip.assert_called_once_with(mock.ANY, ext_net_id)

    def _floatingip_on_port_test_setup(self, hostid):
        router = {'id': 'foo_router_id', 'distributed': True}
        floatingip = {
            'id': _uuid(),
            'port_id': _uuid(),
            'router_id': 'foo_router_id',
        }
        if hostid is not None:
            floatingip['host'] = hostid
        else:
            hostid = 'not_my_host_id'
        routers = {
            'foo_router_id': router
        }
        fipagent = {
            'id': _uuid()
        }

        # NOTE: mock.patch is not needed here since self.mixin is created fresh
        # for each test.  It doesn't work with some methods since the mixin is
        # tested in isolation (e.g. _get_agent_by_type_and_host).
        self.mixin._get_dvr_service_port_hostid = mock.Mock(
            return_value=hostid)
        self.mixin._get_agent_by_type_and_host = mock.Mock(
            return_value=fipagent)
        self.mixin._get_fip_agent_gw_ports = mock.Mock(
            return_value='fip_interface')
        agent = mock.Mock()
        agent.id = fipagent['id']
        self.mixin._process_floating_ips_dvr(self.ctx, routers, [floatingip],
                                             hostid, agent)
        return (router, floatingip)

    def test_floatingip_on_port_no_host_key(self):
        router, fip = self._floatingip_on_port_test_setup(None)

        self.assertNotIn(const.FLOATINGIP_KEY, router)

    def test_floatingip_on_port_with_host(self):
        router, fip = self._floatingip_on_port_test_setup(_uuid())

        self.assertIn(const.FLOATINGIP_KEY, router)
        self.assertIn(fip, router[const.FLOATINGIP_KEY])

    def _setup_test_create_floatingip(self, fip, floatingip_db, router_db):
        port = {
            'id': '1234',
            portbindings.HOST_ID: 'myhost',
            'network_id': 'external_net'
        }

        with mock.patch.object(self.mixin, 'get_router') as grtr,\
                mock.patch.object(self.mixin,
                                  '_get_dvr_service_port_hostid') as vmp,\
                mock.patch.object(
                    self.mixin,
                    '_get_dvr_migrating_service_port_hostid'
                                 ) as mvmp,\
                mock.patch.object(
                    self.mixin,
                    'create_fip_agent_gw_port_if_not_exists') as c_fip,\
                mock.patch.object(l3_dvr_db.l3_db.L3_NAT_db_mixin,
                                  '_update_fip_assoc'):
            grtr.return_value = router_db
            vmp.return_value = 'my-host'
            mvmp.return_value = 'my-future-host'
            registry.publish(
                resources.FLOATING_IP, events.AFTER_CREATE, self,
                payload=events.DBEventPayload(
                    context=mock.Mock(), states=(dict(
                        router_id=router_db['id'], port_id=port['id'],
                        fixed_ip_address='1.2.3.4', id=fip['id'],
                        floating_network_id=fip['floating_network_id']),),
                    resource_id=fip['id'],
                    metadata=dict(association_event=True)))
            return c_fip

    def test_create_floatingip_agent_gw_port_with_dvr_router(self):
        floatingip = {
            'id': _uuid(),
            'router_id': 'foo_router_id'
        }
        router = {'id': 'foo_router_id', 'distributed': True}
        fip = {
            'id': _uuid(),
            'floating_network_id': _uuid(),
            'port_id': _uuid()
        }
        create_fip = (
            self._setup_test_create_floatingip(
                fip, floatingip, router))
        self.assertTrue(create_fip.called)

    def test_create_fip_agent_gw_port_if_not_exists_with_l3_agent(self):
        network_id = _uuid()
        fport_db = {'id': _uuid()}
        self.mixin._get_agent_gw_ports_exist_for_network = mock.Mock(
            return_value=fport_db)

        fipagent = agent_obj.Agent(
                self.ctx,
                id=_uuid(),
                binary='foo-agent',
                host='host',
                agent_type='L3 agent',
                topic='foo_topic',
                configurations={"agent_mode": "dvr_no_external"})
        self.mixin._get_agent_by_type_and_host = mock.Mock(
            return_value=fipagent)
        fport = self.mixin.create_fip_agent_gw_port_if_not_exists(
                                                self.ctx,
                                                network_id,
                                                'host')
        self.assertIsNone(fport)

        fipagent = agent_obj.Agent(
                self.ctx,
                id=_uuid(),
                binary='foo-agent',
                host='host',
                agent_type='L3 agent',
                topic='foo_topic',
                configurations={"agent_mode": "dvr"})
        self.mixin._get_agent_by_type_and_host = mock.Mock(
            return_value=fipagent)
        with mock.patch.object(
            router_obj, "DvrFipGatewayPortAgentBinding",
        ) as dvr_fip_agent_port_obj:
            dvr_fip_agent_port_obj_instance = (
                dvr_fip_agent_port_obj.return_value)
            fport = self.mixin.create_fip_agent_gw_port_if_not_exists(
                                                    self.ctx,
                                                    network_id,
                                                    'host')

        dvr_fip_agent_port_obj_instance.create.assert_not_called()
        self.assertIsNotNone(fport)
        dvr_fip_agent_port_obj_instance.delete.assert_not_called()

    def test_create_fip_agent_gw_port_agent_port_not_created(self):
        network_id = _uuid()
        self.mixin._get_agent_gw_ports_exist_for_network = mock.Mock(
            return_value=None)
        fipagent = agent_obj.Agent(
                self.ctx,
                id=_uuid(),
                binary='foo-agent',
                host='host',
                agent_type='L3 agent',
                topic='foo_topic',
                configurations={"agent_mode": "dvr"})
        self.mixin._get_agent_by_type_and_host = mock.Mock(
            return_value=fipagent)

        with mock.patch.object(
            router_obj, "DvrFipGatewayPortAgentBinding",
        ) as dvr_fip_agent_port_obj,\
            mock.patch.object(
                plugin_utils, "create_port", return_value=None):

            dvr_fip_agent_port_obj_instance = (
                dvr_fip_agent_port_obj.return_value)

            self.assertRaises(
                exceptions.BadRequest,
                self.mixin.create_fip_agent_gw_port_if_not_exists,
                self.ctx, network_id, 'host')

        dvr_fip_agent_port_obj_instance.create.assert_called_once_with()
        self.mixin._get_agent_gw_ports_exist_for_network.\
            assert_called_once_with(
                self.ctx, network_id, 'host', fipagent['id'])
        dvr_fip_agent_port_obj_instance.delete.assert_called_once_with()

    def test_create_fip_agent_gw_port_if_not_exists_duplicate_port(self):
        network_id = _uuid()
        fport_db = {'id': _uuid()}
        self.mixin._get_agent_gw_ports_exist_for_network = mock.Mock(
            side_effect=[None, fport_db])
        fipagent = agent_obj.Agent(
                self.ctx,
                id=_uuid(),
                binary='foo-agent',
                host='host',
                agent_type='L3 agent',
                topic='foo_topic',
                configurations={"agent_mode": "dvr"})
        self.mixin._get_agent_by_type_and_host = mock.Mock(
            return_value=fipagent)

        with mock.patch.object(
            router_obj.DvrFipGatewayPortAgentBinding, 'create',
            side_effect=o_exc.NeutronDbObjectDuplicateEntry(
                mock.Mock(), mock.Mock())
        ) as dvr_fip_gateway_port_agent_binding_create:
            fport = self.mixin.create_fip_agent_gw_port_if_not_exists(
                                                    self.ctx,
                                                    network_id,
                                                    'host')
        dvr_fip_gateway_port_agent_binding_create.assert_called_once_with()
        self.mixin._get_agent_gw_ports_exist_for_network.assert_has_calls([
            mock.call(self.ctx, network_id, 'host', fipagent['id']),
            mock.call(self.ctx, network_id, 'host', fipagent['id'])])
        self.assertIsNotNone(fport)

    def test_create_fip_agent_gw_port_agent_binding_exists(self):
        network_id = _uuid()
        fport_db = {'id': _uuid()}
        self.mixin._get_agent_gw_ports_exist_for_network = mock.Mock(
            side_effect=[None, None])
        fipagent = agent_obj.Agent(
                self.ctx,
                id=_uuid(),
                binary='foo-agent',
                host='host',
                agent_type='L3 agent',
                topic='foo_topic',
                configurations={"agent_mode": "dvr"})
        self.mixin._get_agent_by_type_and_host = mock.Mock(
            return_value=fipagent)
        self.mixin._populate_mtu_and_subnets_for_ports = mock.Mock()

        with mock.patch.object(
            router_obj.DvrFipGatewayPortAgentBinding, 'create',
            side_effect=o_exc.NeutronDbObjectDuplicateEntry(
                mock.Mock(), mock.Mock())
        ) as dvr_fip_gateway_port_agent_binding_create,\
            mock.patch.object(
                plugin_utils, "create_port", return_value=fport_db):
            fport = self.mixin.create_fip_agent_gw_port_if_not_exists(
                                                    self.ctx,
                                                    network_id,
                                                    'host')
        dvr_fip_gateway_port_agent_binding_create.assert_called_once_with()
        self.mixin._get_agent_gw_ports_exist_for_network.assert_has_calls([
            mock.call(self.ctx, network_id, 'host', fipagent['id']),
            mock.call(self.ctx, network_id, 'host', fipagent['id'])])
        self.mixin._populate_mtu_and_subnets_for_ports.assert_has_calls([
            mock.call(self.ctx, [fport_db])])
        self.assertIsNotNone(fport)

    def test_create_floatingip_agent_gw_port_with_non_dvr_router(self):
        floatingip = {
            'id': _uuid(),
            'router_id': 'foo_router_id'
        }
        router = {'id': 'foo_router_id', 'distributed': False}
        fip = {
            'id': _uuid(),
            'floating_network_id': _uuid(),
            'port_id': _uuid()
        }
        create_fip = (
            self._setup_test_create_floatingip(
                fip, floatingip, router))
        self.assertFalse(create_fip.called)

    def test_get_ext_nets_by_host(self):
        ports = [mock.Mock(id=_uuid()) for _ in range(3)]
        fips = [mock.Mock(fixed_port_id=p.id, floating_network_id=_uuid())
                for p in ports]
        expected_ext_nets = set([fip.floating_network_id for fip in fips])
        with mock.patch.object(
            port_obj.Port, 'get_ports_by_host',
            return_value=[p.id for p in ports]
        ) as get_ports_by_host, mock.patch.object(
            self.mixin, '_get_floatingips_by_port_id', return_value=fips
        ) as get_floatingips_by_port_id:
            self.assertEqual(
                expected_ext_nets,
                self.mixin._get_ext_nets_by_host(self.ctx, 'host'))
            get_ports_by_host.assert_called_once_with(self.ctx, 'host')
            get_floatingips_by_port_id.assert_has_calls(
                [mock.call(self.ctx, p.id) for p in ports])

    def _test_create_fip_agent_gw_ports(self, agent_type, agent_mode=None):
        agent = {
            'id': _uuid(),
            'host': 'host',
            'agent_type': agent_type,
            'configurations': {'agent_mode': agent_mode}}
        payload = events.DBEventPayload(
            self.ctx, states=(agent,), resource_id=agent['id'])

        ext_nets = ['ext-net-1', 'ext-net-2']
        with mock.patch.object(
            self.mixin,
            'create_fip_agent_gw_port_if_not_exists'
        ) as create_fip_gw, mock.patch.object(
            self.mixin, "_get_ext_nets_by_host",
            return_value=ext_nets
        ) as get_ext_nets_by_host:

            registry.publish(resources.AGENT, events.AFTER_CREATE, mock.Mock(),
                             payload=payload)

            if agent_type == 'L3 agent' and agent_mode in ['dvr', 'dvr_snat']:
                get_ext_nets_by_host.assert_called_once_with(
                    mock.ANY, 'host')
                create_fip_gw.assert_has_calls(
                    [mock.call(mock.ANY, ext_net, 'host') for
                        ext_net in ext_nets])
            else:
                get_ext_nets_by_host.assert_not_called()
                create_fip_gw.assert_not_called()

    def test_create_fip_agent_gw_ports(self):
        self._test_create_fip_agent_gw_ports(
            agent_type='L3 agent', agent_mode='dvr')
        self._test_create_fip_agent_gw_ports(
            agent_type='L3 agent', agent_mode='dvr_snat')

    def test_create_fip_agent_gw_ports_dvr_no_external_agent(self):
        self._test_create_fip_agent_gw_ports(
            agent_type='L3 agent', agent_mode='dvr_no_external')

    def test_create_fip_agent_gw_ports_non_dvr_agent(self):
        self._test_create_fip_agent_gw_ports(
            agent_type='L3 agent', agent_mode='legacy')

    def test_create_fip_agent_gw_ports_deleted_non_l3_agent(self):
        self._test_create_fip_agent_gw_ports('Other agent type')

    def _test_delete_fip_agent_gw_ports(self, agent_type, agent_mode=None):
        agent = agent_obj.Agent(
            self.ctx, id=_uuid(), agent_type=agent_type,
            configurations={"agent_mode": agent_mode})
        payload = events.DBEventPayload(
            self.ctx, states=(agent,), resource_id=agent.id)

        gw_port = {'id': _uuid(), 'network_id': _uuid()}
        with mock.patch.object(
            self.mixin, '_get_agent_gw_ports',
            return_value=[gw_port]
        ) as get_agent_gw_ports, mock.patch.object(
            self.core_plugin, 'delete_port'
        ) as delete_port:
            registry.publish(resources.AGENT, events.AFTER_DELETE, mock.Mock(),
                             payload=payload)

            if agent_type == 'L3 agent' and agent_mode in ['dvr', 'dvr_snat']:
                get_agent_gw_ports.assert_called_once_with(payload.context,
                                                           agent['id'])
                delete_port.assert_called_once_with(payload.context,
                                                    gw_port['id'])
            else:
                get_agent_gw_ports.assert_not_called()
                delete_port.assert_not_called()

    def test_delete_fip_agent_gw_ports(self):
        self._test_delete_fip_agent_gw_ports(
            agent_type='L3 agent', agent_mode='dvr')
        self._test_delete_fip_agent_gw_ports(
            agent_type='L3 agent', agent_mode='dvr_snat')

    def test_delete_fip_agent_gw_ports_dvr_no_external_agent(self):
        self._test_delete_fip_agent_gw_ports(
            agent_type='L3 agent', agent_mode='dvr_no_external')

    def test_delete_fip_agent_gw_ports_non_dvr_agent(self):
        self._test_delete_fip_agent_gw_ports(
            agent_type='L3 agent', agent_mode='legacy')

    def test_delete_fip_agent_gw_ports_deleted_non_l3_agent(self):
        self._test_delete_fip_agent_gw_ports('Other agent type')

    def _test_update_router_gw_info_external_network_change(self):
        router_dict = {'name': 'test_router', 'admin_state_up': True,
                       'distributed': True}
        router = self._create_router(router_dict)
        with self.network() as net_ext_1,\
                self.network() as net_ext_2,\
                self.subnet() as subnet:
            ext_net_1_id = net_ext_1['network']['id']
            self.core_plugin.update_network(
                self.ctx, ext_net_1_id,
                {'network': {'router:external': True}})
            self.mixin.update_router(
                self.ctx, router['id'],
                {'router': {'external_gateway_info':
                            {'network_id': ext_net_1_id}}})
            self.mixin.add_router_interface(self.ctx, router['id'],
                {'subnet_id': subnet['subnet']['id']})

            ext_net_2_id = net_ext_2['network']['id']
            self.core_plugin.update_network(
                self.ctx, ext_net_2_id,
                {'network': {'router:external': True}})
            self.mixin.update_router(
                self.ctx, router['id'],
                {'router': {'external_gateway_info':
                            {'network_id': ext_net_2_id}}})

            csnat_filters = {'device_owner': [const.DEVICE_OWNER_ROUTER_SNAT]}
            csnat_ports = self.core_plugin.get_ports(
                self.ctx, filters=csnat_filters)
            self.assertEqual(1, len(csnat_ports))

    @mock.patch('neutron.db.l3_dvr_db.is_admin_state_down_necessary',
                return_value=True)
    def test_update_router_gw_info_external_network_change_mocked(self,
            mock_arg):
        # call test with admin_state_down_before_update ENABLED
        self._test_update_router_gw_info_external_network_change()

    @mock.patch('neutron.db.l3_dvr_db.is_admin_state_down_necessary',
                return_value=False)
    def test_update_router_gw_info_external_network_change(self, mock_arg):
        # call test with admin_state_down_before_update DISABLED
        self._test_update_router_gw_info_external_network_change()

    def _test_csnat_ports_removal(self, ha=False):
        router_dict = {'name': 'test_router', 'admin_state_up': True,
                       'distributed': True}
        router = self._create_router(router_dict)
        with self.network() as net_ext,\
                self.subnet() as subnet:
            ext_net_id = net_ext['network']['id']
            self.core_plugin.update_network(
                self.ctx, ext_net_id,
                {'network': {'router:external': True}})
            self.mixin.update_router(
                self.ctx, router['id'],
                {'router': {'external_gateway_info':
                            {'network_id': ext_net_id}}})
            self.mixin.add_router_interface(self.ctx, router['id'],
                {'subnet_id': subnet['subnet']['id']})

            csnat_filters = {'device_owner':
                             [const.DEVICE_OWNER_ROUTER_SNAT]}
            csnat_ports = self.core_plugin.get_ports(
                self.ctx, filters=csnat_filters)
            self.assertEqual(1, len(csnat_ports))

            self.mixin.update_router(
                self.ctx, router['id'],
                {'router': {'admin_state_up': False}})
            self.mixin.update_router(
                self.ctx, router['id'],
                {'router': {'distributed': False, 'ha': ha}})

            csnat_ports = self.core_plugin.get_ports(
                self.ctx, filters=csnat_filters)
            self.assertEqual(0, len(csnat_ports))

    def test_distributed_to_centralized_csnat_ports_removal(self):
        self._test_csnat_ports_removal()

    def test_distributed_to_ha_csnat_ports_removal(self):
        self._test_csnat_ports_removal(ha=True)

    def test_update_router_gw_info_csnat_ports_add(self):
        router_dict = {'name': 'test_router',
                       'admin_state_up': True,
                       'distributed': True}
        router = self._create_router(router_dict)
        with self.network() as net_ext,\
                self.network() as net_int,\
                self.subnet(
                    network=net_int,
                    cidr='2001:db8:1::/64',
                    gateway_ip='2001:db8:1::1',
                    ip_version=const.IP_VERSION_6) as v6_subnet1,\
                self.subnet(
                    network=net_int,
                    cidr='2001:db8:2::/64',
                    gateway_ip='2001:db8:2::1',
                    ip_version=const.IP_VERSION_6) as v6_subnet2,\
                self.subnet(
                    network=net_int,
                    cidr='10.10.10.0/24') as v4_subnet:

            self.core_plugin.update_network(
                self.ctx, net_ext['network']['id'],
                {'network': {'router:external': True}})

            # Add router interface, then set router gateway
            self.mixin.add_router_interface(self.ctx, router['id'],
                {'subnet_id': v6_subnet1['subnet']['id']})
            self.mixin.add_router_interface(self.ctx, router['id'],
                {'subnet_id': v6_subnet2['subnet']['id']})
            self.mixin.add_router_interface(self.ctx, router['id'],
                {'subnet_id': v4_subnet['subnet']['id']})

            dvr_filters = {'device_owner':
                           [const.DEVICE_OWNER_DVR_INTERFACE]}
            dvr_ports = self.core_plugin.get_ports(
                self.ctx, filters=dvr_filters)
            # One for IPv4, one for two IPv6 subnets
            self.assertEqual(2, len(dvr_ports))

            self.mixin.update_router(
                self.ctx, router['id'],
                {'router': {'external_gateway_info':
                            {'network_id': net_ext['network']['id']}}})

            csnat_filters = {'device_owner':
                             [const.DEVICE_OWNER_ROUTER_SNAT]}
            csnat_ports = self.core_plugin.get_ports(
                self.ctx, filters=csnat_filters)
            # One for IPv4, one for two IPv6 subnets
            self.assertEqual(2, len(csnat_ports))

            # Remove v4 subnet interface from router
            self.mixin.remove_router_interface(
                self.ctx, router['id'],
                {'subnet_id': v4_subnet['subnet']['id']})

            dvr_ports = self.core_plugin.get_ports(
                self.ctx, filters=dvr_filters)
            self.assertEqual(1, len(dvr_ports))

            csnat_ports = self.core_plugin.get_ports(
                self.ctx, filters=csnat_filters)
            self.assertEqual(1, len(csnat_ports))
            self.assertEqual(2, len(csnat_ports[0]['fixed_ips']))

    def _test_update_router_interface_port_ip_not_allowed(self, device_owner):
        router, subnet_v4, subnet_v6 = self._setup_router_with_v4_and_v6()
        device_filter = {'device_owner': [device_owner]}
        ports = self.core_plugin.get_ports(self.ctx, filters=device_filter)
        self.assertRaises(
            exceptions.BadRequest,
            self.core_plugin.update_port,
            self.ctx, ports[0]['id'],
            {'port': {'fixed_ips': [
                {'ip_address': "20.0.0.100",
                 'subnet_id': subnet_v4['subnet']['id']},
                {'ip_address': "20.0.0.101",
                 'subnet_id': subnet_v4['subnet']['id']}]}})

    def test_update_router_centralized_snat_port_ip_not_allowed(self):
        self._test_update_router_interface_port_ip_not_allowed(
            const.DEVICE_OWNER_ROUTER_SNAT)

    def test_update_router_interface_distributed_port_ip_not_allowed(self):
        self._test_update_router_interface_port_ip_not_allowed(
            const.DEVICE_OWNER_DVR_INTERFACE)

    def test_remove_router_interface_csnat_ports_removal(self):
        router_dict = {'name': 'test_router', 'admin_state_up': True,
                       'distributed': True}
        router = self._create_router(router_dict)
        with self.network() as net_ext,\
                self.subnet() as subnet1,\
                self.subnet(cidr='20.0.0.0/24') as subnet2:
            ext_net_id = net_ext['network']['id']
            self.core_plugin.update_network(
                self.ctx, ext_net_id,
                {'network': {'router:external': True}})
            self.mixin.update_router(
                self.ctx, router['id'],
                {'router': {'external_gateway_info':
                            {'network_id': ext_net_id}}})
            self.mixin.add_router_interface(self.ctx, router['id'],
                {'subnet_id': subnet1['subnet']['id']})
            self.mixin.add_router_interface(self.ctx, router['id'],
                {'subnet_id': subnet2['subnet']['id']})

            csnat_filters = {'device_owner':
                             [const.DEVICE_OWNER_ROUTER_SNAT]}
            csnat_ports = self.core_plugin.get_ports(
                self.ctx, filters=csnat_filters)
            self.assertEqual(2, len(csnat_ports))

            dvr_filters = {'device_owner':
                           [const.DEVICE_OWNER_DVR_INTERFACE]}
            dvr_ports = self.core_plugin.get_ports(
                self.ctx, filters=dvr_filters)
            self.assertEqual(2, len(dvr_ports))

            self.mixin.remove_router_interface(
                self.ctx, router['id'], {'port_id': dvr_ports[0]['id']})

            csnat_ports = self.core_plugin.get_ports(
                self.ctx, filters=csnat_filters)
            self.assertEqual(1, len(csnat_ports))
            self.assertEqual(dvr_ports[1]['fixed_ips'][0]['subnet_id'],
                             csnat_ports[0]['fixed_ips'][0]['subnet_id'])

            dvr_ports = self.core_plugin.get_ports(
                self.ctx, filters=dvr_filters)
            self.assertEqual(1, len(dvr_ports))

    def _setup_router_with_v4_and_v6(self):
        router_dict = {'name': 'test_router', 'admin_state_up': True,
                       'distributed': True}
        router = self._create_router(router_dict)
        with self.network() as net_ext, self.network() as net_int:
            ext_net_id = net_ext['network']['id']
            self.core_plugin.update_network(
                self.ctx, ext_net_id,
                {'network': {'router:external': True}})
            self.mixin.update_router(
                self.ctx, router['id'],
                {'router': {'external_gateway_info':
                            {'network_id': ext_net_id}}})
            with self.subnet(
                network=net_int, cidr='20.0.0.0/24') as subnet_v4,\
                self.subnet(network=net_int, cidr='fe80::/64',
                            gateway_ip='fe80::1', ip_version=const.IP_VERSION_6
                            ) as subnet_v6:
                self.mixin.add_router_interface(self.ctx, router['id'],
                    {'subnet_id': subnet_v4['subnet']['id']})
                self.mixin.add_router_interface(self.ctx, router['id'],
                    {'subnet_id': subnet_v6['subnet']['id']})
                return router, subnet_v4, subnet_v6

    def test_undo_router_interface_change_on_csnat_error(self):
        self._test_undo_router_interface_change_on_csnat_error(False)

    def test_undo_router_interface_change_on_csnat_error_revert_failure(self):
        self._test_undo_router_interface_change_on_csnat_error(True)

    def _test_undo_router_interface_change_on_csnat_error(self, fail_revert):
        router, subnet_v4, subnet_v6 = self._setup_router_with_v4_and_v6()
        net = {'network': {'id': subnet_v6['subnet']['network_id'],
                           'tenant_id': subnet_v6['subnet']['tenant_id']}}
        orig_update = self.mixin._core_plugin.update_port

        def update_port(*args, **kwargs):
            # 1st port update is the interface, 2nd is csnat, 3rd is revert
            # we want to simulate errors after the 1st
            update_port.calls += 1
            if update_port.calls == 2:
                raise RuntimeError('csnat update failure')
            if update_port.calls == 3 and fail_revert:
                # this is to ensure that if the revert fails, the original
                # exception is raised (not this ValueError)
                raise ValueError('failure from revert')
            return orig_update(*args, **kwargs)
        update_port.calls = 0
        self.mixin._core_plugin.update_port = update_port

        with self.subnet(network=net, cidr='fe81::/64',
                         gateway_ip='fe81::1', ip_version=const.IP_VERSION_6
                         ) as subnet2_v6:
            self.mixin.add_router_interface(self.ctx, router['id'],
                {'subnet_id': subnet2_v6['subnet']['id']})
            if fail_revert:
                # a revert failure will mean the interface is still added
                # so we can't re-add it
                return
            # starting over should work if first interface was cleaned up
            self.mixin.add_router_interface(self.ctx, router['id'],
                {'subnet_id': subnet2_v6['subnet']['id']})

    def test_remove_router_interface_csnat_ports_removal_with_ipv6(self):
        router, subnet_v4, subnet_v6 = self._setup_router_with_v4_and_v6()
        csnat_filters = {'device_owner':
                         [const.DEVICE_OWNER_ROUTER_SNAT]}
        csnat_ports = self.core_plugin.get_ports(
            self.ctx, filters=csnat_filters)
        self.assertEqual(2, len(csnat_ports))
        dvr_filters = {'device_owner':
                       [const.DEVICE_OWNER_DVR_INTERFACE]}
        dvr_ports = self.core_plugin.get_ports(
            self.ctx, filters=dvr_filters)
        self.assertEqual(2, len(dvr_ports))
        self.mixin.remove_router_interface(
            self.ctx, router['id'],
            {'subnet_id': subnet_v4['subnet']['id']})
        csnat_ports = self.core_plugin.get_ports(
            self.ctx, filters=csnat_filters)
        self.assertEqual(1, len(csnat_ports))
        self.assertEqual(
            subnet_v6['subnet']['id'],
            csnat_ports[0]['fixed_ips'][0]['subnet_id'])

        dvr_ports = self.core_plugin.get_ports(
            self.ctx, filters=dvr_filters)
        self.assertEqual(1, len(dvr_ports))

    def _test__validate_router_migration_notify_advanced_services(self):
        router = {'name': 'foo_router', 'admin_state_up': False}
        router_db = self._create_router(router)
        with mock.patch.object(l3_dvr_db.registry, 'publish') as mock_publish:
            self.mixin._validate_router_migration(
                self.ctx, router_db, {'distributed': True})
            mock_publish.assert_called_once_with(
                'router', 'before_update', self.mixin,
                payload=mock.ANY)

            payload = mock_publish.call_args_list[0][1]['payload']
            self.assertEqual(self.ctx, payload.context)
            self.assertEqual(router_db, payload.latest_state)

    def _assert_mock_called_with_router(self, mock_fn, router_id):
        router = mock_fn.call_args[1].get('router_db')
        self.assertEqual(router_id, router.id)

    def test__validate_router_migration_notify_advanced_services_mocked(self):
        # call test with admin_state_down_before_update ENABLED
        self._test__validate_router_migration_notify_advanced_services()

    def test__validate_router_migration_notify_advanced_services(self):
        # call test with admin_state_down_before_update DISABLED
        self._test__validate_router_migration_notify_advanced_services()

    def test_validate_add_router_interface_by_subnet_notify_advanced_services(
            self):
        router = {'name': 'foo_router', 'admin_state_up': False}
        router_db = self._create_router(router)
        with self.network() as net, \
                self.subnet(network={'network': net['network']}) as sub, \
                mock.patch.object(
                    self.mixin,
                    '_notify_attaching_interface') as mock_notify:
            interface_info = {'subnet_id': sub['subnet']['id']}
            self.mixin.add_router_interface(self.ctx, router_db.id,
                                            interface_info)
            # NOTE(slaweq): here we are just checking if mock_notify was called
            # with kwargs which we are expecting, but we can't check exactly if
            # router_db was object which we are expecting and because of that
            # below we are checking if router_db used as argument in
            # mock_notify call is the has same id as the one which we are
            # expecting
            mock_notify.assert_called_once_with(self.ctx, router_db=mock.ANY,
                                                port=mock.ANY,
                                                interface_info=interface_info)
            self._assert_mock_called_with_router(mock_notify, router_db.id)

    def test_validate_add_router_interface_by_port_notify_advanced_services(
            self):
        router = {'name': 'foo_router', 'admin_state_up': False}
        router_db = self._create_router(router)
        with self.network() as net, \
                self.subnet(network={'network': net['network']}) as sub, \
                self.port(subnet=sub) as port, \
                mock.patch.object(
                    self.mixin,
                    '_notify_attaching_interface') as mock_notify:
            interface_info = {'port_id': port['port']['id']}
            self.mixin.add_router_interface(self.ctx, router_db.id,
                                            interface_info)
            # NOTE(slaweq): here we are just checking if mock_notify was called
            # with kwargs which we are expecting, but we can't check exactly if
            # router_db was object which we are expecting and because of that
            # below we are checking if router_db used as argument in
            # mock_notify call is the has same id as the one which we are
            # expecting.
            mock_notify.assert_called_once_with(self.ctx, router_db=mock.ANY,
                                                port=mock.ANY,
                                                interface_info=interface_info)
            self._assert_mock_called_with_router(mock_notify, router_db.id)

    def test__generate_arp_table_and_notify_agent(self):
        fixed_ip = {
            'ip_address': '1.2.3.4',
            'subnet_id': _uuid()}
        mac_address = "00:11:22:33:44:55"
        expected_arp_table = {
            'ip_address': fixed_ip['ip_address'],
            'subnet_id': fixed_ip['subnet_id'],
            'mac_address': mac_address}
        notifier = mock.Mock()
        ports = [{'id': _uuid(), 'device_id': 'router_1'},
                 {'id': _uuid(), 'device_id': 'router_2'}]
        with mock.patch.object(self.core_plugin, "get_ports",
                               return_value=ports):
            self.mixin._generate_arp_table_and_notify_agent(
                self.ctx, fixed_ip, mac_address, notifier)
        notifier.assert_has_calls([
            mock.call(self.ctx, "router_1", expected_arp_table),
            mock.call(self.ctx, "router_2", expected_arp_table)])

    def _test_update_arp_entry_for_dvr_service_port(
            self, device_owner, action):
        router_dict = {'name': 'test_router', 'admin_state_up': True,
                       'distributed': True}
        router = self._create_router(router_dict)
        plugin = mock.Mock()
        directory.add_plugin(plugin_constants.CORE, plugin)
        l3_notify = self.mixin.l3_rpc_notifier = mock.Mock()
        port = {
            'id': 'my_port_id',
            'fixed_ips': [
                {'subnet_id': '51edc9e0-24f9-47f2-8e1e-2a41cb691323',
                 'ip_address': '10.0.0.11'},
                {'subnet_id': '2b7c8a07-6f8e-4937-8701-f1d5da1a807c',
                 'ip_address': '10.0.0.21'},
                {'subnet_id': '48534187-f077-4e81-93ff-81ec4cc0ad3b',
                 'ip_address': 'fd45:1515:7e0:0:f816:3eff:fe1a:1111'}],
            'mac_address': 'my_mac',
            'device_owner': device_owner
        }
        dvr_port = {
            'id': 'dvr_port_id',
            'fixed_ips': mock.ANY,
            'device_owner': const.DEVICE_OWNER_DVR_INTERFACE,
            'device_id': router['id']
        }
        plugin.get_ports.return_value = [dvr_port]
        if action == 'add':
            self.mixin.update_arp_entry_for_dvr_service_port(
                self.ctx, port)
            self.assertEqual(3, l3_notify.add_arp_entry.call_count)
        elif action == 'del':
            self.mixin.delete_arp_entry_for_dvr_service_port(
                self.ctx, port)
            self.assertEqual(3, l3_notify.del_arp_entry.call_count)

    def test_update_arp_entry_for_dvr_service_port_added(self):
        action = 'add'
        device_owner = const.DEVICE_OWNER_LOADBALANCER
        self._test_update_arp_entry_for_dvr_service_port(device_owner, action)

    def test_update_arp_entry_for_dvr_service_port_deleted(self):
        action = 'del'
        device_owner = const.DEVICE_OWNER_LOADBALANCER
        self._test_update_arp_entry_for_dvr_service_port(device_owner, action)

    def test_add_router_interface_csnat_ports_failure(self):
        router_dict = {'name': 'test_router', 'admin_state_up': True,
                       'distributed': True}
        router = self._create_router(router_dict)
        with self.network() as net_ext,\
                self.subnet() as subnet:
            ext_net_id = net_ext['network']['id']
            self.core_plugin.update_network(
                self.ctx, ext_net_id,
                {'network': {'router:external': True}})
            self.mixin.update_router(
                self.ctx, router['id'],
                {'router': {'external_gateway_info':
                            {'network_id': ext_net_id}}})
            with mock.patch.object(self.mixin,
                                   '_add_csnat_router_interface_port') as f:
                f.side_effect = RuntimeError()
                self.assertRaises(
                    l3_exc.RouterInterfaceAttachmentConflict,
                    self.mixin.add_router_interface,
                    self.ctx, router['id'],
                    {'subnet_id': subnet['subnet']['id']})
                filters = {
                    'device_id': [router['id']],
                }
                router_ports = self.core_plugin.get_ports(self.ctx, filters)
                self.assertEqual(1, len(router_ports))
                self.assertEqual(const.DEVICE_OWNER_ROUTER_GW,
                                 router_ports[0]['device_owner'])

    def test_csnat_port_not_created_on_RouterPort_update_exception(self):
        router_dict = {'name': 'test_router', 'admin_state_up': True,
                       'distributed': True}
        router = self._create_router(router_dict)
        with self.network() as net_ext,\
                self.subnet() as subnet:
            ext_net_id = net_ext['network']['id']
            self.core_plugin.update_network(
                self.ctx, ext_net_id,
                {'network': {'router:external': True}})
            self.mixin.update_router(
                self.ctx, router['id'],
                {'router': {'external_gateway_info':
                            {'network_id': ext_net_id}}})
            net_id = subnet['subnet']['network_id']
            with mock.patch.object(router_obj.RouterPort,
                                   'create') as rtrport_update:
                rtrport_update.side_effect = Exception()
                self.assertRaises(
                    l3_exc.RouterInterfaceAttachmentConflict,
                    self.mixin.add_router_interface,
                    self.ctx, router['id'],
                    {'subnet_id': subnet['subnet']['id']})
                filters = {
                    'network_id': [net_id],
                    'device_owner': [const.DEVICE_OWNER_ROUTER_SNAT]
                }
                router_ports = self.core_plugin.get_ports(self.ctx, filters)
                self.assertEqual(0, len(router_ports))

    def test_add_router_interface_by_port_failure(self):
        router_dict = {'name': 'test_router',
                       'admin_state_up': True,
                       'distributed': True}
        router = self._create_router(router_dict)
        with self.subnet(cidr='10.10.10.0/24') as subnet:
            port_dict = {
                'device_id': '',
                'device_owner': '',
                'admin_state_up': True,
                'fixed_ips': [{'subnet_id': subnet['subnet']['id'],
                               'ip_address': '10.10.10.100'}]
            }
            net_id = subnet['subnet']['network_id']
            port_res = self.create_port(net_id, port_dict)
            port = self.deserialize(self.fmt, port_res)
            self.assertIn('port', port, message='Create port failed.')

            orig_update_port = self.mixin._core_plugin.update_port
            call_info = {'count': 0}

            def _fake_update_port(*args, **kwargs):
                call_info['count'] += 1
                if call_info['count'] == 2:
                    raise RuntimeError()
                else:
                    return orig_update_port(*args, **kwargs)

            # NOTE(trananhkma): expect that update_port() only raises an error
            # at the 2nd function call (Update owner after actual process
            # again in order).
            with mock.patch.object(self.mixin._core_plugin, 'update_port',
                                   side_effect=_fake_update_port):
                self.assertRaises(
                    RuntimeError,
                    self.mixin.add_router_interface,
                    self.ctx, router['id'], {'port_id': port['port']['id']})
            # expire since we are re-using the session which might have stale
            # ports in it
            self.ctx.session.expire_all()
            port_info = self.core_plugin.get_port(self.ctx, port['port']['id'])
            self.assertEqual(port_dict['device_id'], port_info['device_id'])
            self.assertEqual(port_dict['device_owner'],
                             port_info['device_owner'])

    def test__get_sync_routers_check_gw_port_host(self):
        router_dict = {'name': 'test_router', 'admin_state_up': True,
                       'distributed': True}
        router = self._create_router(router_dict)
        with self.network() as public,\
                self.subnet() as subnet:
            ext_net_1_id = public['network']['id']
            self.core_plugin.update_network(
                self.ctx, ext_net_1_id,
                {'network': {'router:external': True}})
            self.mixin.update_router(
                self.ctx, router['id'],
                {'router': {'external_gateway_info':
                            {'network_id': ext_net_1_id}}})
            self.mixin.add_router_interface(self.ctx, router['id'],
                {'subnet_id': subnet['subnet']['id']})
            routers = self.mixin._get_sync_routers(self.ctx,
                                                   router_ids=[router['id']])
            self.assertIsNone(routers[0]['gw_port_host'])

            agent = mock.Mock()
            agent.host = "fake-host"
            bind = mock.Mock()
            bind.l3_agent_id = "fake-id"
            with mock.patch.object(
                rb_obj.RouterL3AgentBinding, 'get_objects',
                return_value=[bind]), mock.patch.object(
                    agent_obj.Agent, 'get_object',
                    return_value=agent):
                routers = self.mixin._get_sync_routers(
                    self.ctx, router_ids=[router['id']])
                self.assertEqual("fake-host", routers[0]['gw_port_host'])

    def test_is_router_distributed(self):
        router_id = 'router_id'
        with mock.patch.object(self.mixin, 'get_router') as \
                mock_get_router:
            mock_get_router.return_value = {'distributed': True}
            self.assertTrue(
                self.mixin.is_router_distributed(self.ctx, router_id))

    @mock.patch.object(l3_dvr_db, "is_port_bound")
    def test_get_ports_under_dvr_connected_subnet(self, is_port_bound_mock):
        router_dict = {'name': 'test_router', 'admin_state_up': True,
                       'distributed': True}
        router = self._create_router(router_dict)
        with self.network() as network,\
                self.subnet(network=network) as subnet:
            self.mixin._core_plugin.get_allowed_address_pairs_for_ports = (
                mock.Mock())
            fake_bound_ports_ids = []

            def fake_is_port_bound(port):
                return port['id'] in fake_bound_ports_ids

            is_port_bound_mock.side_effect = fake_is_port_bound

            for _ in range(4):
                port_res = self.create_port(
                    network['network']['id'],
                    {'fixed_ips': [{'subnet_id': subnet['subnet']['id']}]})
                port_id = self.deserialize(self.fmt, port_res)['port']['id']
                if len(fake_bound_ports_ids) < 2:
                    fake_bound_ports_ids.append(port_id)

            self.mixin.add_router_interface(self.ctx, router['id'],
                {'subnet_id': subnet['subnet']['id']})
            dvr_subnet_ports = self.mixin.get_ports_under_dvr_connected_subnet(
                self.ctx, subnet['subnet']['id'])
            dvr_subnet_ports_ids = [p['id'] for p in dvr_subnet_ports]
            self.assertCountEqual(fake_bound_ports_ids, dvr_subnet_ports_ids)
            (self.mixin._core_plugin.get_allowed_address_pairs_for_ports.
                assert_called_once_with(self.ctx, dvr_subnet_ports_ids))

    @mock.patch.object(plugin_utils, 'can_port_be_bound_to_virtual_bridge',
                       return_value=True)
    def test__get_assoc_data_valid_vnic_type(self, *args):
        with mock.patch.object(self.mixin, '_internal_fip_assoc_data') as \
                mock_fip_assoc_data, \
                mock.patch.object(self.mixin, '_get_router_for_floatingip') \
                as mock_router_fip, \
                mock.patch.object(self.mixin, 'is_router_distributed',
                                  return_value=True):
            port = {portbindings.VNIC_TYPE: portbindings.VNIC_NORMAL}
            mock_fip_assoc_data.return_value = (port, 'subnet_id', 'ip_addr')
            mock_router_fip.return_value = 'router_id'
            fip = {'port_id': 'port_id'}
            self.assertEqual(
                ('port_id', 'ip_addr', 'router_id'),
                self.mixin._get_assoc_data(self.ctx, fip, mock.Mock()))

    @mock.patch.object(plugin_utils, 'can_port_be_bound_to_virtual_bridge',
                       return_value=False)
    def test__get_assoc_data_invalid_vnic_type(self, *args):
        with mock.patch.object(self.mixin, '_internal_fip_assoc_data') as \
                mock_fip_assoc_data, \
                mock.patch.object(self.mixin, '_get_router_for_floatingip') \
                as mock_router_fip, \
                mock.patch.object(self.mixin, 'is_router_distributed',
                                  return_value=True):
            port = {portbindings.VNIC_TYPE: portbindings.VNIC_NORMAL}
            mock_fip_assoc_data.return_value = (port, 'subnet_id', 'ip_addr')
            mock_router_fip.return_value = 'router_id'
            self.assertRaises(
                exceptions.BadRequest,
                self.mixin._get_assoc_data, self.ctx, mock.ANY, mock.Mock())

    def test__delete_dvr_internal_ports(self):
        payload = mock.Mock()
        payload.context = mock.Mock()
        payload.latest_state = {'distributed': True}
        payload.metadata = {'new_network_id': 'fake-net-1',
                            'network_id': 'fake-net-2'}
        plugin = mock.Mock()
        directory.add_plugin(plugin_constants.CORE, plugin)
        plugin.get_ports.return_value = []
        with mock.patch.object(self.mixin,
                               'delete_floatingip_agent_gateway_port') as \
            del_port, \
            mock.patch.object(
                self.mixin.l3_rpc_notifier,
                'delete_fipnamespace_for_ext_net') as \
            del_fip_ns, \
            mock.patch.object(router_obj.DvrFipGatewayPortAgentBinding,
                              "delete_objects") as del_binding:
            self.mixin._delete_dvr_internal_ports(
                None, None, resources.ROUTER_GATEWAY, payload)
            del_port.assert_called_once()
            del_fip_ns.assert_called_once()
            del_binding.assert_called_once()
